# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for OpenAI models."""

from typing import Any
import unittest
from unittest import mock

import langfun.core as lf
from langfun.core import modalities as lf_modalities
from langfun.core.llms import openai_compatible
import pyglove as pg
import requests


def mock_chat_completion_request(url: str, json: dict[str, Any], **kwargs):
  del url, kwargs
  messages = json['messages']
  if len(messages) > 1:
    system_message = f' system={messages[0]["content"]}'
  else:
    system_message = ''

  if 'response_format' in json:
    response_format = f' format={json["response_format"]["type"]}'
  else:
    response_format = ''

  choices = []
  for k in range(json['n']):
    if json.get('logprobs'):
      logprobs = dict(
          content=[
              dict(
                  token='chosen_token',
                  logprob=0.5,
                  top_logprobs=[
                      dict(
                          token=f'alternative_token_{i + 1}',
                          logprob=0.1
                      ) for i in range(3)
                  ]
              )
          ]
      )
    else:
      logprobs = None

    choices.append(dict(
        message=dict(
            content=(
                f'Sample {k} for message.{system_message}{response_format}'
            )
        ),
        logprobs=logprobs,
    ))
  response = requests.Response()
  response.status_code = 200
  response._content = pg.to_json_str(
      dict(
          choices=choices,
          usage=lf.LMSamplingUsage(
              prompt_tokens=100,
              completion_tokens=100,
              total_tokens=200,
          ),
      )
  ).encode()
  return response


def mock_chat_completion_request_vision(
    url: str, json: dict[str, Any], **kwargs
):
  del url, kwargs
  choices = []
  urls = [
      c['image_url']['url']
      for c in json['messages'][0]['content'] if c['type'] == 'image_url'
  ]
  for k in range(json['n']):
    choices.append(pg.Dict(
        message=pg.Dict(
            content=f'Sample {k} for message: {"".join(urls)}'
        ),
        logprobs=None,
    ))
  response = requests.Response()
  response.status_code = 200
  response._content = pg.to_json_str(
      dict(
          choices=choices,
          usage=lf.LMSamplingUsage(
              prompt_tokens=100,
              completion_tokens=100,
              total_tokens=200,
          ),
      )
  ).encode()
  return response


class OpenAIComptibleTest(unittest.TestCase):
  """Tests for OpenAI compatible language model."""

  def test_request_args(self):
    self.assertEqual(
        openai_compatible.OpenAICompatible(
            api_endpoint='https://test-server',
            model='test-model'
        )._request_args(
            lf.LMSamplingOptions(
                temperature=1.0, stop=['\n'], n=1, random_seed=123
            )
        ),
        dict(
            model='test-model',
            top_logprobs=None,
            n=1,
            temperature=1.0,
            stop=['\n'],
            seed=123,
        ),
    )

  def test_call_chat_completion(self):
    with mock.patch('requests.Session.post') as mock_request:
      mock_request.side_effect = mock_chat_completion_request
      lm = openai_compatible.OpenAICompatible(
          api_endpoint='https://test-server', model='test-model',
      )
      self.assertEqual(
          lm('hello', sampling_options=lf.LMSamplingOptions(n=2)),
          'Sample 0 for message.',
      )

  def test_call_chat_completion_with_logprobs(self):
    with mock.patch('requests.Session.post') as mock_request:
      mock_request.side_effect = mock_chat_completion_request
      lm = openai_compatible.OpenAICompatible(
          api_endpoint='https://test-server', model='test-model',
      )
      results = lm.sample(['hello'], logprobs=True)
      for result in results:
        result.rebind(
            {path: 0 for path in pg.query(result, '.*total_call_interval')},
            skip_notification=True,
        )
      self.assertEqual(len(results), 1)
      self.assertEqual(
          results[0],
          lf.LMSamplingResult(
              [
                  lf.LMSample(
                      response=lf.AIMessage(
                          text='Sample 0 for message.',
                          metadata={
                              'score': 0.0,
                              'logprobs': [(
                                  'chosen_token',
                                  0.5,
                                  [
                                      ('alternative_token_1', 0.1),
                                      ('alternative_token_2', 0.1),
                                      ('alternative_token_3', 0.1),
                                  ],
                              )],
                              'is_cached': False,
                              'usage': lf.LMSamplingUsage(
                                  prompt_tokens=100,
                                  completion_tokens=100,
                                  total_tokens=200,
                                  estimated_cost=None,
                              ),
                          },
                          tags=['lm-response'],
                      ),
                      logprobs=[(
                          'chosen_token',
                          0.5,
                          [
                              ('alternative_token_1', 0.1),
                              ('alternative_token_2', 0.1),
                              ('alternative_token_3', 0.1),
                          ],
                      )],
                  )
              ],
              usage=lf.LMSamplingUsage(
                  prompt_tokens=100,
                  completion_tokens=100,
                  total_tokens=200,
                  estimated_cost=None,
              ),
          ),
      )

  def test_call_chat_completion_vision(self):

    class FakeImage(lf_modalities.Image):

      @property
      def mime_type(self) -> str:
        return 'image/png'

    with mock.patch('requests.Session.post') as mock_request:
      mock_request.side_effect = mock_chat_completion_request_vision
      lm_1 = openai_compatible.OpenAICompatible(
          api_endpoint='https://test-server',
          model='test-model1',
      )
      lm_2 = openai_compatible.OpenAICompatible(
          api_endpoint='https://test-server',
          model='test-model2',
      )
      for lm in (lm_1, lm_2):
        self.assertEqual(
            lm(
                lf.UserMessage(
                    'hello <<[[image]]>>',
                    image=FakeImage.from_uri('https://fake/image')
                ),
                sampling_options=lf.LMSamplingOptions(n=2)
            ),
            'Sample 0 for message: https://fake/image',
        )

    class TextOnlyModel(openai_compatible.OpenAICompatible):

      class ModelInfo(lf.ModelInfo):
        input_modalities: list[str] = lf.ModelInfo.TEXT_INPUT_ONLY

      @property
      def model_info(self) -> lf.ModelInfo:
        return TextOnlyModel.ModelInfo('text-only-model')

    lm_3 = TextOnlyModel(
        api_endpoint='https://test-server', model='test-model3'
    )
    with self.assertRaisesRegex(ValueError, 'Unsupported modality'):
      lm_3(
          lf.UserMessage(
              'hello <<[[image]]>>',
              image=FakeImage.from_uri('https://fake/image')
          ),
      )

  def test_sample_chat_completion(self):
    with mock.patch('requests.Session.post') as mock_request:
      mock_request.side_effect = mock_chat_completion_request
      lm = openai_compatible.OpenAICompatible(
          api_endpoint='https://test-server', model='test-model'
      )
      results = lm.sample(
          ['hello', 'bye'], sampling_options=lf.LMSamplingOptions(n=3)
      )
      for result in results:
        result.rebind(
            {path: 0 for path in pg.query(result, '.*total_call_interval')},
            skip_notification=True,
        )

    self.assertEqual(len(results), 2)
    self.assertEqual(
        results[0],
        lf.LMSamplingResult(
            [
                lf.LMSample(
                    lf.AIMessage(
                        'Sample 0 for message.',
                        score=0.0,
                        logprobs=None,
                        is_cached=False,
                        usage=lf.LMSamplingUsage(
                            prompt_tokens=33,
                            completion_tokens=33,
                            total_tokens=66,
                            estimated_cost=None,
                        ),
                        tags=[lf.Message.TAG_LM_RESPONSE],
                    ),
                    score=0.0,
                    logprobs=None,
                ),
                lf.LMSample(
                    lf.AIMessage(
                        'Sample 1 for message.',
                        score=0.0,
                        logprobs=None,
                        is_cached=False,
                        usage=lf.LMSamplingUsage(
                            prompt_tokens=33,
                            completion_tokens=33,
                            total_tokens=66,
                            estimated_cost=None,
                        ),
                        tags=[lf.Message.TAG_LM_RESPONSE],
                    ),
                    score=0.0,
                    logprobs=None,
                ),
                lf.LMSample(
                    lf.AIMessage(
                        'Sample 2 for message.',
                        score=0.0,
                        logprobs=None,
                        is_cached=False,
                        usage=lf.LMSamplingUsage(
                            prompt_tokens=33,
                            completion_tokens=33,
                            total_tokens=66,
                            estimated_cost=None,
                        ),
                        tags=[lf.Message.TAG_LM_RESPONSE],
                    ),
                    score=0.0,
                    logprobs=None,
                ),
            ],
            usage=lf.LMSamplingUsage(
                prompt_tokens=100, completion_tokens=100, total_tokens=200,
                estimated_cost=None,
            ),
        ),
    )
    self.assertEqual(
        results[1],
        lf.LMSamplingResult(
            [
                lf.LMSample(
                    lf.AIMessage(
                        'Sample 0 for message.',
                        score=0.0,
                        logprobs=None,
                        is_cached=False,
                        usage=lf.LMSamplingUsage(
                            prompt_tokens=33,
                            completion_tokens=33,
                            total_tokens=66,
                            estimated_cost=None,
                        ),
                        tags=[lf.Message.TAG_LM_RESPONSE],
                    ),
                    score=0.0,
                    logprobs=None,
                ),
                lf.LMSample(
                    lf.AIMessage(
                        'Sample 1 for message.',
                        score=0.0,
                        logprobs=None,
                        is_cached=False,
                        usage=lf.LMSamplingUsage(
                            prompt_tokens=33,
                            completion_tokens=33,
                            total_tokens=66,
                            estimated_cost=None,
                        ),
                        tags=[lf.Message.TAG_LM_RESPONSE],
                    ),
                    score=0.0,
                    logprobs=None,
                ),
                lf.LMSample(
                    lf.AIMessage(
                        'Sample 2 for message.',
                        score=0.0,
                        logprobs=None,
                        is_cached=False,
                        usage=lf.LMSamplingUsage(
                            prompt_tokens=33,
                            completion_tokens=33,
                            total_tokens=66,
                            estimated_cost=None,
                        ),
                        tags=[lf.Message.TAG_LM_RESPONSE],
                    ),
                    score=0.0,
                    logprobs=None,
                ),
            ],
            usage=lf.LMSamplingUsage(
                prompt_tokens=100, completion_tokens=100, total_tokens=200,
                estimated_cost=None,
            ),
        ),
    )

  def test_sample_with_contextual_options(self):
    with mock.patch('requests.Session.post') as mock_request:
      mock_request.side_effect = mock_chat_completion_request
      lm = openai_compatible.OpenAICompatible(
          api_endpoint='https://test-server', model='test-model'
      )
      with lf.use_settings(sampling_options=lf.LMSamplingOptions(n=2)):
        results = lm.sample(['hello'])
      for result in results:
        result.rebind(
            {path: 0 for path in pg.query(result, '.*total_call_interval')},
            skip_notification=True,
        )

    self.assertEqual(len(results), 1)
    self.assertEqual(
        results[0],
        lf.LMSamplingResult(
            [
                lf.LMSample(
                    lf.AIMessage(
                        'Sample 0 for message.',
                        score=0.0,
                        logprobs=None,
                        is_cached=False,
                        usage=lf.LMSamplingUsage(
                            prompt_tokens=50,
                            completion_tokens=50,
                            total_tokens=100,
                        ),
                        tags=[lf.Message.TAG_LM_RESPONSE],
                    ),
                    score=0.0,
                    logprobs=None,
                ),
                lf.LMSample(
                    lf.AIMessage(
                        'Sample 1 for message.',
                        score=0.0,
                        logprobs=None,
                        is_cached=False,
                        usage=lf.LMSamplingUsage(
                            prompt_tokens=50,
                            completion_tokens=50,
                            total_tokens=100,
                        ),
                        tags=[lf.Message.TAG_LM_RESPONSE],
                    ),
                    score=0.0,
                    logprobs=None,
                ),
            ],
            usage=lf.LMSamplingUsage(
                prompt_tokens=100, completion_tokens=100, total_tokens=200
            ),
        )
    )

  def test_call_with_system_message(self):
    with mock.patch('requests.Session.post') as mock_request:
      mock_request.side_effect = mock_chat_completion_request
      lm = openai_compatible.OpenAICompatible(
          api_endpoint='https://test-server', model='test-model'
      )
      self.assertEqual(
          lm(
              lf.UserMessage(
                  'hello',
                  system_message=lf.SystemMessage('hi'),
              ),
              sampling_options=lf.LMSamplingOptions(n=2)
          ),
          '''Sample 0 for message. system=[{'type': 'text', 'text': 'hi'}]''',
      )

  def test_call_with_json_schema(self):
    with mock.patch('requests.Session.post') as mock_request:
      mock_request.side_effect = mock_chat_completion_request
      lm = openai_compatible.OpenAICompatible(
          api_endpoint='https://test-server', model='test-model'
      )
      self.assertEqual(
          lm(
              lf.UserMessage(
                  'hello',
                  json_schema={
                      'type': 'object',
                      'properties': {
                          'name': {'type': 'string'},
                      },
                      'required': ['name'],
                      'title': 'Person',
                  }
              ),
              sampling_options=lf.LMSamplingOptions(n=2)
          ),
          'Sample 0 for message. format=json_schema',
      )

    # Test bad json schema.
    with self.assertRaisesRegex(ValueError, '`json_schema` must be a dict'):
      lm(lf.UserMessage('hello', json_schema='foo'))

    with self.assertRaisesRegex(
        ValueError, 'The root of `json_schema` must have a `title` field'
    ):
      lm(lf.UserMessage('hello', json_schema={}))

  def test_call_with_context_limit_error(self):
    def mock_context_limit_error(*args, **kwargs):
      del args, kwargs
      response = requests.Response()
      response.status_code = 400
      response._content = b'string_above_max_length'
      return response

    with mock.patch('requests.Session.post') as mock_request:
      mock_request.side_effect = mock_context_limit_error
      lm = openai_compatible.OpenAICompatible(
          api_endpoint='https://test-server', model='test-model'
      )
      with self.assertRaisesRegex(
          lf.ContextLimitError, 'string_above_max_length'
      ):
        lm(lf.UserMessage('hello'))


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, Tuple
import unittest

from langfun.assistant.capabilities.gui import bounding_box_parser
from langfun.assistant.capabilities.gui import location


class BoundingBoxTest(unittest.TestCase):

  def assert_bbox_equal(
      self,
      expected: Dict[str, Tuple[int, int, int, int]],
      actual: Dict[str, location.BBox],
  ):
    self.assertEqual(len(expected), len(actual))
    for key, value in expected.items():
      self.assertIn(key, actual)
      self.assertEqual(value[0], actual[key].x)
      self.assertEqual(value[1], actual[key].y)
      self.assertEqual(value[2], actual[key].right)
      self.assertEqual(value[3], actual[key].bottom)

  def test_bbox_basic_functionality(self):
    bbox = bounding_box_parser.GeminiBBox(xmin=10, ymin=20, xmax=50, ymax=80)

    self.assertEqual(bbox.xmin, 10)
    self.assertEqual(bbox.ymin, 20)
    self.assertEqual(bbox.xmax, 50)
    self.assertEqual(bbox.ymax, 80)

    # Test the `scale` method
    target_size = (800, 600)
    source_size = (1000, 1000)
    scaled_bbox = bbox.scale(target_size, source_size)

    self.assertEqual(scaled_bbox.xmin, 8)
    self.assertEqual(scaled_bbox.ymin, 12)
    self.assertEqual(scaled_bbox.xmax, 40)
    self.assertEqual(scaled_bbox.ymax, 48)

    # Test the `resize` method
    resized_bbox = bbox.resize((1200, 800), 2)

    self.assertEqual(resized_bbox.xmin, 0)
    self.assertEqual(resized_bbox.ymin, 0)
    self.assertEqual(resized_bbox.xmax, 70)
    self.assertEqual(resized_bbox.ymax, 110)

    # Test the `to_gui_bbox` method
    gui_bbox = bbox.to_gui_bbox()
    self.assertEqual(gui_bbox.x, 10)
    self.assertEqual(gui_bbox.y, 20)
    self.assertEqual(gui_bbox.right, 50)
    self.assertEqual(gui_bbox.bottom, 80)

  def test_simple_json(self):
    json_text = '{"search button": [10, 20, 100, 200]}'
    expected = {'search button': (16, 6, 160, 60)}
    result = bounding_box_parser.parse_and_convert_json(
        json_text, screen_size=(800, 600)
    )
    self.assert_bbox_equal(expected, result)

  def test_multiple_objects(self):
    json_text = (
        '{"button1": [10, 20, 100, 200], "button2": [30, 40, 130, 240]}'
    )
    expected = {'button1': (16, 6, 160, 60), 'button2': (32, 18, 192, 78)}
    result = bounding_box_parser.parse_and_convert_json(
        json_text, screen_size=(800, 600)
    )
    self.assert_bbox_equal(expected, result)

  def test_nested_json(self):
    json_text = (
        '{"buttons": {"search": [10, 20, 100, 200], "cancel": [30, 40, 130,'
        ' 240]}}'
    )
    expected = {'search': (16, 6, 160, 60), 'cancel': (32, 18, 192, 78)}
    result = bounding_box_parser.parse_and_convert_json(
        json_text, screen_size=(800, 600)
    )
    self.assert_bbox_equal(expected, result)

  def test_json_in_code_block(self):
    json_text = '```\n{"search button": [10, 20, 100, 200]}\n```'
    expected = {'search button': (16, 6, 160, 60)}
    result = bounding_box_parser.parse_and_convert_json(
        json_text, screen_size=(800, 600)
    )
    self.assert_bbox_equal(expected, result)

  def test_extract_json_candidate_from_text(self):
    test_cases = [
        (
            'Some text before ```json\n{"key": "value"}\n``` and after',
            '{"key": "value"}',
        ),
        (
            'Some text before ```\n{"key": "value"}\n``` and after',
            '{"key": "value"}',
        ),
        ('{"key": "value"}', '{"key": "value"}'),
        (
            '```json\n{\n  "name": "Test",\n  "version": 1\n}\n```',
            '{\n  "name": "Test",\n  "version": 1\n}',
        ),
        (
            '```\n{\n  "name": "Test",\n  "version": 1\n}\n```',
            '{\n  "name": "Test",\n  "version": 1\n}',
        ),
        (
            '   ```json\n   {"spaced_json": true}   \n```   ',
            '{"spaced_json": true}',
        ),
        (
            'No code block here, just plain text.',
            'No code block here, just plain text.',
        ),
        (
            '```JSON\n{"case_test": "uppercase_json_tag"}\n```',
            '{"case_test": "uppercase_json_tag"}',
        ),
        ('', ''),
        ('   ', ''),
        (
            (
                'First block: ```json\n{"first": true}\n``` Second block:'
                ' ```json\n{"second": false}\n```'
            ),
            '{"first": true}',
        ),
        (
            (
                'First block: ```\n{"first_code": true}\n``` Second block:'
                ' ```\n{"second_code": false}\n```'
            ),
            '{"first_code": true}',
        ),
        (
            '```json  \n  {"leading_trailing_space_in_block": "test"}  \n  ```',
            '{"leading_trailing_space_in_block": "test"}',
        ),
        (
            (
                '```\n  {"leading_trailing_space_in_block_no_json_tag": "test"}'
                '  \n  ```'
            ),
            '{"leading_trailing_space_in_block_no_json_tag": "test"}',
        ),
    ]

    for raw_text, expected_json_str in test_cases:
      with self.subTest(raw_text=raw_text):
        self.assertEqual(
            bounding_box_parser.extract_json_candidate_from_text(raw_text),
            expected_json_str,
        )

  def test_dict_in_list(self):
    json_text = '```json\n[\n  {"box_2d": [61, 22, 160, 95]}\n]\n```'
    expected = {'box_2d': (22, 61, 95, 160)}
    result = bounding_box_parser.parse_and_convert_json(json_text)
    self.assert_bbox_equal(expected, result)

  def test_dict_with_label(self):
    json_text = """```json
[
  {"box_2d": [634, 416, 820, 482], "label": "the inner vertical side of the rightmost lower protrusion of the green polygon"},
  {"box_2d": [820, 328, 872, 352], "label": "the purple number '1' located to its left"}
]
```"""
    expected = {'box_2d': (328, 820, 352, 872)}
    result = bounding_box_parser.parse_and_convert_json(json_text)
    print('result: ', result)
    self.assert_bbox_equal(expected, result)

  def test_invalid_json(self):
    json_text = 'This is not a valid JSON'
    result = bounding_box_parser.parse_and_convert_json(json_text)
    self.assertEqual({}, result)

  def test_empty_input(self):
    json_text = ''
    result = bounding_box_parser.parse_and_convert_json(json_text)
    self.assertEqual({}, result)

  def test_invalid_list_length(self):
    # Test with a list of length 3
    json_text = '{"button": [10, 20, 100]}'
    result = bounding_box_parser.parse_and_convert_json(json_text)
    self.assertEqual({}, result)

    # Test with a list of length 5
    json_text = '{"button": [10, 20, 100, 200, 300]}'
    result = bounding_box_parser.parse_and_convert_json(json_text)
    self.assertEqual({}, result)

    # Test with an empty list
    json_text = '{"button": []}'
    result = bounding_box_parser.parse_and_convert_json(json_text)
    self.assertEqual({}, result)

  def test_list_input(self):
    json_text = '[10, 20, 100, 200]'
    expected = {'element': (20, 10, 200, 100)}
    result = bounding_box_parser.parse_and_convert_json(json_text)
    self.assert_bbox_equal(expected, result)

  def test_default_screen_size(self):
    json_text = '{"button": [10, 20, 100, 200]}'
    expected = {'button': (20, 10, 200, 100)}
    result = bounding_box_parser.parse_and_convert_json(json_text)
    self.assert_bbox_equal(expected, result)

  def test_float_numbers(self):
    json_text = '{"button": [10.5, 20.2, 100.7, 200.9]}'
    expected = {'button': (20, 10, 200, 100)}  # Expected integer coordinates
    result = bounding_box_parser.parse_and_convert_json(
        json_text, screen_size=(1000, 1000)
    )
    self.assert_bbox_equal(expected, result)

  def test_type_error_handling(self):
    json_text = '{"button": ["text", 20, 100, 200]}'
    result = bounding_box_parser.parse_and_convert_json(
        json_text, screen_size=(1000, 1000)
    )
    self.assertEqual({}, result)
    json_text = '{"button": [None, 20, 100, 200]}'
    result = bounding_box_parser.parse_and_convert_json(
        json_text, screen_size=(1000, 1000)
    )
    self.assertEqual({}, result)

  def test_malformed_json(self):
    # Missing quotes around keys
    json_text = '{search button: [10, 20, 100, 200]}'
    result = bounding_box_parser.parse_and_convert_json(json_text)
    self.assertEqual({}, result)

    # Unbalanced brackets
    json_text = '{"search button": [10, 20, 100, 200}'
    result = bounding_box_parser.parse_and_convert_json(json_text)
    self.assertEqual({}, result)

    # Incorrect comma usage
    json_text = '{"search button", [10, 20, 100, 200]}'
    result = bounding_box_parser.parse_and_convert_json(json_text)
    self.assertEqual({}, result)

  def test_mixed_data_types(self):
    # String values in coordinates
    json_text = '{"button": ["10", "20", "100", "200"]}'
    expected = {'button': (20, 10, 200, 100)}
    result = bounding_box_parser.parse_and_convert_json(
        json_text, screen_size=(1000, 1000)
    )
    self.assert_bbox_equal(expected, result)

  def test_none_values(self):
    json_text = '{"button": [null, 20, 100, 200]}'
    result = bounding_box_parser.parse_and_convert_json(
        json_text, screen_size=(1000, 1000)
    )
    self.assertEqual({}, result)

    json_text = '{"button": [10, null, 100, 200]}'
    result = bounding_box_parser.parse_and_convert_json(
        json_text, screen_size=(1000, 1000)
    )
    self.assertEqual({}, result)

  def test_large_coordinates(self):
    json_text = '{"button": [10000, 20000, 100000, 200000]}'
    expected = {'button': (20000, 10000, 200000, 100000)}
    result = bounding_box_parser.parse_and_convert_json(
        json_text, screen_size=(1000, 1000)
    )
    self.assert_bbox_equal(expected, result)

  def test_different_string_formats(self):
    # Newlines
    json_text = '{\n"search button": [10, 20, 100, 200]\n}'
    expected = {'search button': (16, 6, 160, 60)}
    result = bounding_box_parser.parse_and_convert_json(
        json_text, screen_size=(800, 600)
    )
    self.assert_bbox_equal(expected, result)

  def test_deeply_nested_json(self):
    json_text = (
        '{"layer1": {"layer2": {"layer3": {"button": [10, 20, 100, 200]}}}}'
    )
    expected = {'button': (20, 10, 200, 100)}
    result = bounding_box_parser.parse_and_convert_json(json_text)
    self.assert_bbox_equal(expected, result)

if __name__ == '__main__':
  unittest.main()



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
import time
import unittest

from langfun.core.eval.v2 import progress as progress_lib
import pyglove as pg

Progress = progress_lib.Progress


class ProgressTest(unittest.TestCase):

  def test_basic(self):
    p = Progress()
    self.assertFalse(p.is_started)
    self.assertFalse(p.is_stopped)
    self.assertFalse(p.is_completed)
    self.assertFalse(p.is_skipped)
    self.assertFalse(p.is_failed)
    self.assertEqual(p.num_completed, 0)
    self.assertIsNone(p.elapse)
    self.assertIsNone(p.start_time_str)
    self.assertIsNone(p.stop_time_str)
    self.assertTrue(math.isnan(float(p)))

    p.start(10)
    self.assertEqual(p.num_total, 10)
    self.assertTrue(p.is_started)
    self.assertFalse(p.is_stopped)
    self.assertIsNotNone(p.start_time_str)
    self.assertIsNotNone(p.elapse)
    self.assertEqual(float(p), 0.0)

    with pg.views.html.controls.HtmlControl.track_scripts() as scripts:
      p.increment_processed()
      p.increment_failed()
      p.increment_skipped()
    # Does not triggger scripts as progress is not rendered yet.
    self.assertEqual(len(scripts), 0)
    self.assertEqual(p.num_completed, 3)
    self.assertIn(
        '3/10',
        p.to_html(extra_flags=dict(interactive=True)).content,
    )
    with pg.views.html.controls.HtmlControl.track_scripts() as scripts:
      p.increment_processed()
      p.increment_failed()
      p.increment_skipped()
    self.assertEqual(len(scripts), 24)
    self.assertEqual(p.num_completed, 6)
    self.assertIn(
        '6/10',
        p.to_html(extra_flags=dict(interactive=False)).content,
    )
    p.update_execution_summary(
        dict(
            evaluate=pg.object_utils.TimeIt.Status(name='evaluate', elapse=1.0)
        )
    )
    p.stop()
    elapse1 = p.elapse
    time.sleep(0.1)
    self.assertEqual(p.elapse, elapse1)
    self.assertTrue(p.is_stopped)
    self.assertIsNotNone(p.stop_time_str)


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Structures for location objects in an image."""

import unittest
import langfun as lf
from langfun.assistant.capabilities.gui import drawing
from langfun.assistant.capabilities.gui import location
from PIL import Image as pil_image


class DrawingTest(unittest.TestCase):

  def test_blank_image(self):
    image = drawing.blank_image((3, 3), (0, 0, 0))
    self.assertIsInstance(image, lf.Image)
    self.assertEqual(image.size, (3, 3))
    self.assertEqual(image.image_format, 'png')
    self.assertIn(b'\x89PNG\r\n\x1a\n', image.to_bytes())
    image2 = drawing.blank_image((3, 3), (0, 0, 0), pil=True)
    self.assertIsInstance(image2, pil_image.Image)
    image2 = lf.Image.from_pil_image(image2)
    self.assertIsInstance(image2, lf.Image)
    self.assertEqual(image2.to_bytes(), image.to_bytes())

  def test_draw_bboxes(self):
    image = drawing.blank_image((5, 5), (0, 0, 0))
    image2 = drawing.draw_bboxes(image, [])
    self.assertIs(image, image2)

    image = drawing.draw_bboxes(
        image,
        [
            location.BBox(0, 0, 3, 3),
            location.BBox(1, 1, 4, 4),
        ],
        line_color='red',
        line_width=1,
    )
    self.assertEqual(image.size, (5, 5))
    self.assertIsInstance(image, lf.Image)

  def test_draw_points(self):
    image = drawing.blank_image((5, 5), (0, 0, 0))
    image2 = drawing.draw_points(image, [])
    self.assertIs(image, image2)

    image = drawing.draw_points(
        image,
        [
            location.Coordinate(1, 1),
            location.Coordinate(2, 2),
        ],
        color='red',
        radius=1,
    )
    self.assertEqual(image.size, (5, 5))
    self.assertIsInstance(image, lf.Image)

  def test_draw_calibration_lines(self):
    image = drawing.blank_image((5, 5), (0, 0, 0))
    image = drawing.draw_calibration_lines(
        image,
        location.Coordinate(2, 2),
        vline_color='red',
        vline_width=1,
        hline_color='green',
        hline_width=1,
    )
    self.assertEqual(image.size, (5, 5))
    self.assertIsInstance(image, lf.Image)

  def test_draw_cursor(self):
    image = drawing.blank_image((50, 50), (255, 0, 0))
    image = drawing.draw_cursor(image, location.Coordinate(10, 10))
    self.assertEqual(image.size, (50, 50))
    self.assertIsInstance(image, lf.Image)

  def test_draw_ref_points(self):
    image = drawing.blank_image((50, 50), (255, 0, 0))
    image = drawing.draw_ref_points(image, 10, 10, 3, 3)
    self.assertEqual(image.size, (50, 50))
    self.assertIsInstance(image, lf.Image)

  def test_draw_text(self):
    image = drawing.blank_image((50, 50), (255, 0, 0))
    image = drawing.draw_text(image, 'Hello World', location.Coordinate(10, 10))
    self.assertEqual(image.size, (50, 50))
    self.assertIsInstance(image, lf.Image)

if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common langfun templates."""

# pylint: disable=g-bad-import-order
# pylint: disable=g-importing-member

from langfun.core.structured.schema import include_method_in_prompt

from langfun.core.structured.schema import Missing
from langfun.core.structured.schema import MISSING
from langfun.core.structured.schema import Unknown
from langfun.core.structured.schema import UNKNOWN

from langfun.core.structured.schema import Schema
from langfun.core.structured.schema import SchemaProtocol
from langfun.core.structured.schema import schema_spec

from langfun.core.structured.schema import SchemaError
from langfun.core.structured.schema import JsonError

from langfun.core.structured.schema import class_dependencies
from langfun.core.structured.schema import class_definition
from langfun.core.structured.schema import class_definitions
from langfun.core.structured.schema import annotation
from langfun.core.structured.schema import structure_from_python

from langfun.core.structured.schema import schema_repr
from langfun.core.structured.schema import source_form
from langfun.core.structured.schema import value_repr

from langfun.core.structured.schema_generation import generate_class
from langfun.core.structured.schema_generation import classgen_example
from langfun.core.structured.schema_generation import default_classgen_examples

from langfun.core.structured.function_generation import function_gen

from langfun.core.structured.mapping import Mapping
from langfun.core.structured.mapping import MappingError
from langfun.core.structured.mapping import MappingExample

from langfun.core.structured.parsing import parse
from langfun.core.structured.parsing import aparse
from langfun.core.structured.parsing import call
from langfun.core.structured.parsing import acall

from langfun.core.structured.querying import track_queries
from langfun.core.structured.querying import QueryInvocation

from langfun.core.structured.querying import LfQuery
from langfun.core.structured.querying import query
from langfun.core.structured.querying import aquery
from langfun.core.structured.querying import query_and_reduce
from langfun.core.structured.querying import query_protocol

from langfun.core.structured.querying import query_prompt
from langfun.core.structured.querying import query_output
from langfun.core.structured.querying import query_reward

from langfun.core.structured.description import describe
from langfun.core.structured.completion import complete
from langfun.core.structured.completion import acomplete

from langfun.core.structured.scoring import score
from langfun.core.structured.scoring import ascore

from langfun.core.structured.tokenization import tokenize
from langfun.core.structured.tokenization import atokenize


# Expose default examples for structured operations so users could refer to
# them.
from langfun.core.structured.parsing import default_parse_examples
from langfun.core.structured.description import default_describe_examples

# Default examples.


# pylint: enable=g-importing-member
# pylint: enable=g-bad-import-order



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Conversation history with FIFO support."""

from typing import Annotated, Tuple
import langfun.core as lf


class ConversationHistory(lf.Memory):
  """A memory that keeps tracking of the conversation history."""

  max_turns: Annotated[
      int | None,
      'Max number of conversation turns to keep. If None, there is no limit.',
  ] = None

  def _on_bound(self):
    super()._on_bound()
    self.reset()

  @property
  def turns(self) -> list[Tuple[lf.Message, lf.Message]]:
    """Return conversation turns."""
    return self._turns

  @property
  def messages(self) -> list[lf.Message]:
    """Return conversation messages."""
    messages = []
    for r in self.turns:
      messages.append(r[0])
      messages.append(r[1])
    return messages

  def _recollect(self, **kwargs) -> str:
    return '\n'.join([f'{m.sender}: {m.text}' for m in self.messages])

  def _remember(
      self,
      value: tuple[str | lf.Message, str | lf.Message],
      **kwargs
  ) -> None:
    if self.max_turns and self.max_turns == len(self._turns):
      self._turns.pop(0)
    self._turns.append(
        (lf.UserMessage.from_value(value[0]), lf.AIMessage.from_value(value[1]))
    )

  def _reset(self, **kwargs) -> None:
    self._turns = []



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Evaluation experiment."""

import abc
import datetime
import functools
import hashlib
import inspect
import os
import re
from typing import Annotated, Any, Callable, Literal, Optional

import langfun.core as lf
from langfun.core.eval.v2 import example as example_lib
from langfun.core.eval.v2 import progress as progress_lib
import pyglove as pg


class Experiment(lf.Component, pg.views.HtmlTreeView.Extension):
  """Evaluation Experiment.

  # Experiment Structure.

  An evaluation experiment is structured as a tree of evaluation tasks, where
  each task is represented as a node in the tree. Leaf tasks are instances of
  `Evaluation` with concrete hyper-parameter values. Nodes such as `Suite` and
  `Evaluation` that utilize `pg.oneof` are non-leaf tasks, as they represent
  multiple configurations. Leaf tasks can be retrieved using property
  `leaf_nodes`, while non-leaf tasks can be retrieved using property
  `nonleaf_nodes`. An experiment without any leaf tasks is considered
  empty.

  For example:

    ```
    Suite(
        MyEvaluation1(
            lm=pg.oneof([lm1, lm2]),
        ),
        Suite(
            MyEvaluation2(
                lm=lm1,
            ),
            MyEvaluation3(
                lm=lm2,
            ),
        )
    )
    ```

  In this example:
  - The two `Suite` nodes and the `MyEvaluation1` node (with pg.oneof) are
    non-leaf nodes, as they contain leaf tasks.
  - There are four leaf nodes. Two leaf nodes under `MyEvaluation1`, which
    correspond to `MyEvaluation1` instances with `lm1` and `lm2` as
    hyper-parameters respectively. The objects of `MyEvaluation2` and
    `MyEvaluation3` are also leaf nodes as they have specific hyper-parameter
    values.

  # Running an Experiment

  To run an experiment, users can call `Experiment.run`. This will execute the
  experiment using a specified `Runner` instance (e.g., 'parallel' or
  'sequential'). Progress and results will be periodically written to HTML
  files. Users can also assign an id to each run, which will identify the output
  directory of that run.

  By default, the experiment will resume from the latest run under the root
  directory (using the ID 'latest'). Users can specify 'new' to start a fresh
  run or provide a specific run ID (typically in the format %Y%m%d_%<number>).
  Additionally, when initiating a new run, users may specify a `warm_start_from`
  directory to restore the experiment’s state from a previous run.

  Examples:

    ```
    root_dir = '/path/to/experiment/root'

    # Resume the latest experiment run, or start a new run if none exists.
    experiment.run(root_dir)

    # Equivalent to:
    experiment.run(root_dir, 'latest')

    # Start a new, clean run.
    experiment.run(root_dir, 'new')

    # Start a new run with a warm start from the another run located at
    # '/path/to/another/run' (e.g. /my_expreriment/run_20241031_1).
    experiment.run(root_dir, 'new', warm_start_from='/path/to/another/run')

    # Resume run '20241031_1', re-running failed examples and recomputing
    # metrics as needed.
    experiment.run(root_dir, '20241031_1')

    # Reprocess the previous run located in 'run_20241031_1'.
    experiment.run(root_dir, '20241031_1', reprocess=True)
    ```

  # Experiment Registration and Lookup

  Experiments can be registered by setting a class-level NAME attribute.
  Users can then retrieve a registered experiment using Experiment.find(name).

  For example:

  ```
  class MyEval(lf.eval.v2.Evaluation):
    NAME = 'my_eval'

  class MyEvalVariation1(MyEval):
    NAME = 'my_eval/gemini'
    lm = pg.oneof([lf.llms.GeminiPro(), lf.llms.GeminiFlash(), ...])

  class MyEvalVariation2(MyEval):
    NAME = 'my_eval/openai'
    lm = pg.oneof([lf.llms.Gpt4o(), lf.llms.Gpt4Turbo(), ...])

  # Run all experiments with "gemini" in their name.
  experiment = Experiment.find('.*/gemini')
  experiment.run()

  # Run all experiments with "my_eval" in their name.
  experiment = Experiment.find('my_eval.*')
  experiment.run()
  ```

  # Checkpointing

  Experiments support checkpointing, which is enabled by default. It allows 
  users to resume their experiments from a saved state. When an experiment runs,
  it creates a new directory for that run and saves the current state to a
  checkpoint file. If the experiment is interrupted or fails, users can resume
  it by specifying the 'id' or 'warm_start_from' argument (shown above) to
  seamlessly continue from previously saved state without starting over.

  # Monitoring and Reporting

  Evaluations can take considerable time to complete, so Langfun provides
  several tools to monitor progress. Progress bars display the status of each
  evaluation: HTML-based progress bars update in real time within Colab
  notebooks, while text-based progress bars appear in the terminal using tqdm.

  Additionally, Langfun generates HTML files at regular intervals to provide
  progress updates and detailed evaluation results. These files are saved in
  the evaluation's output directory, organized as follows:

   root_dir>                # Root directory of the experiment.
     |_ <run_id>            # Root directory of current run.
         |_ summary.html    # Summary of the run. Updated every 60 seconds.
         |_ <experiment_cls>      # Directory of a particular experiment type.
             |_ <experiment_hash> # Directory of a particular experiment config.
             |_ index.html        # Experiment report. Updated every 60 seconds.
             |_ 1.html            # Detailed evaluation output of example 1.
             |_ 2.html            # Detailed evaluation output of example 2.
             |_ ...

  # Experiment Plugins

  Experiment can be extended by plugins. Plugins can listen to the events of
  experiment execution and produce additional outputs. For example, a plugin
  can be added to an experiment to generate additional metrics or to save
  additional data to a database. More details will be added in the future.
  """

  #
  # Class-level functionalities.
  #

  # An global unique str as a well-known name for an experiment,
  # which can be retrieved by `Experiment.find(name)`. If None, the experiment
  # does not have a well-known name, thus users need to create the experiment
  # by constructing it explicitly.
  NAME = None

  # Global registry for experiment classes with GLOBAL_ID.
  _NAME_TO_CLASS = {}

  def __init_subclass__(cls):
    super().__init_subclass__()

    if inspect.isabstract(cls):
      return

    if cls.NAME is not None:
      cls._NAME_TO_CLASS[cls.NAME] = cls

  @classmethod
  def find(cls, pattern: str) -> 'Experiment':
    """Finds an experiment by global name.

    Args:
      pattern: A regular expression to match the global names of registered 
        experiments.

    Returns:
      An experiment object. If multiple experiments are found, a
        `Suite` of matched experiments will be returned. If no experiment is
        found, an empty `Suite` will be returned.
    """
    if pattern in cls._NAME_TO_CLASS:
      return cls._NAME_TO_CLASS[pattern]()
    regex = re.compile(pattern)
    selected = []
    for cls_name, exp_cls in cls._NAME_TO_CLASS.items():
      if regex.match(cls_name):
        selected.append(exp_cls())
    return selected[0] if len(selected) == 1 else Suite(selected)

  #
  # Instance-level functionalities.
  #

  progress: Annotated[
      progress_lib.Progress,
      'The progress of the experiment.'
  ] = progress_lib.Progress()

  usage_summary: Annotated[
      lf.UsageSummary,
      'The usage summary of the experiment.'
  ] = lf.UsageSummary()

  plugins: Annotated[
      list['Plugin'],
      (
          'Plugins for current experiment, which can listen to the events '
          'of experiment execution and produce additional outputs.'
      )
  ] = []

  def _on_bound(self):
    super()._on_bound()
    self.__dict__.pop('hash', None)
    self.__dict__.pop('dir', None)
    self._reset()

  #
  # Identity of an experiment.
  #

  @property
  def id(self) -> str:
    """Returns the ID for this evaluaton."""
    return f'{self.__class__.__name__}@{self.hash}'

  def definition(self, hide_default_values: bool = True) -> str:
    """Returns the definition of the experiment."""
    return self.format(
        compact=False,
        hide_default_values=hide_default_values,
        use_inferred=True,
        exclude_keys=('progress', 'usage_summary')
    )

  @functools.cached_property
  def hash(self) -> str:
    """A 8-byte MD5 hash computed from experiment identity."""
    identity = self.format(
        compact=True, hide_default_values=True, use_inferred=True,
        exclude_keys=('plugins', 'progress', 'usage_summary')
    )
    return hashlib.md5(identity.encode()).hexdigest()[:8]

  @classmethod
  def link(cls, path: str) -> str:
    assert path.startswith('/'), path
    return f'file://{path}'

  # Placehold Google-internal methods.

  #
  # Hierarchy of an experiment tree.
  #

  @property
  @abc.abstractmethod
  def children(self) -> list['Experiment']:
    """Returns the child experiments."""

  @property
  @abc.abstractmethod
  def is_leaf(self) -> bool:
    """Returns whether the experiment is a leaf node."""

  def empty(self) -> bool:
    """Returns whether the experiment is empty."""
    return not self.leaf_nodes

  @functools.cached_property
  def nodes(self) -> list['Experiment']:
    """Returns all the experiment nodes in the subtree (including self)."""
    nodes = [self]
    for child in self.children:
      nodes.extend(child.nodes)
    return nodes

  @functools.cached_property
  def leaf_nodes(self) -> list['Experiment']:
    """Returns the leaf nodes.

    The leaf-nodes of an experiment are evaluable objects that has materilized
    hyper-parameters.
    """
    if self.is_leaf:
      return [self]

    nodes = []
    for child in self.children:
      nodes.extend(child.leaf_nodes)
    return nodes

  @functools.cached_property
  def nonleaf_nodes(self) -> list['Experiment']:
    """Returns the non-leaf nodes."""
    if self.is_leaf:
      return []
    nodes = [self]
    for child in self.children:
      nodes.extend(child.nonleaf_nodes)
    return nodes

  @functools.cached_property
  def parent(self) -> Optional['Experiment']:
    """Returns the parent experiment."""
    parent = self.sym_parent
    while parent is not None and not isinstance(parent, Experiment):
      parent = parent.sym_parent
    return parent

  def get(self, evaluation_id: str) -> Optional['Experiment']:
    """Returns the experiment by ID."""
    for leaf in self.leaf_nodes:
      if leaf.id == evaluation_id:
        return leaf
    return None

  #
  # Mutable states during evaluaton.
  #

  def reset(self) -> None:
    """Resets the experiment for a new run."""
    self.progress.reset()
    self.rebind(
        usage_summary=lf.UsageSummary(),
        skip_notification=True,
        raise_on_no_change=False
    )
    if self.is_leaf:
      self._reset()
    else:
      for child in self.children:
        child.reset()

  def _reset(self) -> None:
    """Subclass could override."""

  #
  # Helper methods for running the evaluation without explicitly creating the
  # runner.
  #

  def run(
      self,
      root_dir: str,
      id: str | None = None,   # pylint: disable=redefined-builtin
      *,
      runner: str = 'parallel',
      warm_start_from: str | None = None,
      filter: Callable[['Experiment'], bool] | None = None,   # pylint: disable=redefined-builtin
      example_ids: list[int] | None = None,
      shuffle_inputs: bool = False,
      raise_if_has_error: bool = False,
      reevaluate_upon_previous_errors: bool = True,
      reprocess: bool | list[int] = False,
      generate_example_html: Literal['new', 'all', 'no'] | list[int] = 'new',
      process_timeout: int | None = None,
      use_cache: Literal['global', 'per_dataset', 'no'] = 'per_dataset',
      note: str | None = None,
      tags: list[str] | None = None,
      plugins: list['Plugin'] | None = None,
      **kwargs
  ) -> 'Run':
    """Runs the experiment.

    Examples:
      # Start a new run under root_dir.
      experiment.run(root_dir, 'new')

      # Continue the latest experiment run.
      experiment.run(root_dir, 'latest')

      # Continue the latest experiment run or start a new run if it does not
      # exist.
      experiment.run(root_dir)

      # Start a new run and warm start from another run's directory
      # '/path/to/another/run_20241031_1/'.
      experiment.run(
          root_dir, 'new',
          warm_start_from='/path/to/another/run_20241031_1/'
      )

      # Reprocess previous run under sub-dir 'run_20241031_1'.
      experiment.run(root_dir, '20241031_1', reprocess=True)

    Args:
      root_dir: The root of the output directory of the experiment.
      id: The ID of the current run. It can be None, a special keyword 'latest'
        or 'new', or a datetime string in format `%Y%m%d%_%` (e.g. 20241031_1).
        If None, it will use the latest run ID under the root directory or
        create a new run based on the current time if no previous run exists.
        If `latest`, it will use the latest run ID under the root directory.
        If `new`, it will create a new run ID based on the current time.
      runner: The runner to use. If None, it will use the default runner for
        the experiment.
      warm_start_from: The ID of the previous run to warm start from. If None,
        it will continue the experiment identified by `id` from where it left
        off. Otherwise, it will create a new experiment run by warming start.
      filter: A filter function to decide whether an experiment should be run
        or not.
      example_ids: The example IDs to run. If None, it will run all examples.
      shuffle_inputs: If True, the order of evaluatin examples will be shuffled.
        Neverthless, the example ID remains unchanged for each example.
      raise_if_has_error: If True, it will raise an error if any example fails.
        Otherwise, it will continue and report the error in the output.
      reevaluate_upon_previous_errors: If True, it will reevaluate the example
        if the previous checkpointed run has error when experiment runs with
        existing checkpoints.
      reprocess: A boolean or a list of example IDs. If boolean, it indicates
        that whether all the examples to be evaluated will be reprocessed,
        meaning that existing checkpoints will be ignored. If a list of
        example IDs, it indicates that only the specified examples will be
        reprocessed.
      generate_example_html: Among 'new', 'all', 'no' or a list of example IDs.
        If 'new', generate HTML files for all newly processed examples, and
          keep/copy existing HTML files for unchanged examples.
        If 'all', generate HTML files for all examples.
        If 'no', do not generate HTML files for any examples.
        If a list of example IDs, generate HTML files for the specified
        examples.
      process_timeout: The timeout in seconds for each process. If None, it
        will use the default timeout for the runner.
      use_cache: Whether to use LLM cache for the experiment.
        If `global`, it will use a global cache shared by all experiments.
        If `per_dataset`, it will use a cache dedicated for each dataset.
        If `no`, it will not use any cache.
      note: The note for the current run.
      tags: The tags for the current run.
      plugins: Runner plugins to use.
      **kwargs: Additional kwargs to pass to the runner.

    Returns:
      The current run.
    """
    if plugins is not None:
      kwargs['plugins'] = plugins
    runner = Runner.create(
        runner,
        current_run=Run(
            root_dir=root_dir,
            experiment=pg.Ref(self),
            id=RunId.from_id(id, root_dir),
            warm_start_from=warm_start_from,
            filter=filter,
            example_ids=example_ids,
            shuffle_inputs=shuffle_inputs,
            raise_if_has_error=raise_if_has_error,
            reevaluate_upon_previous_errors=reevaluate_upon_previous_errors,
            reprocess=reprocess,
            generate_example_html=generate_example_html,
            use_cache=use_cache,
            process_timeout=process_timeout,
            note=note,
            tags=tags or [],
        ),
        **kwargs
    )
    runner.run()
    return runner.current_run

  def run_preconfigured(
      self,
      root_dir: str | None = None,
      id: str | None = None,   # pylint: disable=redefined-builtin
      **kwargs
  ) -> 'Run':
    """Runs the experiment with pre-configured kwargs from `cls.RUN_ARGS`.

    This helper method allows users to config running arguments as a part of
    the class.

    Args:
      root_dir: root directory of the experiment.
      id: ID of the current run.
      **kwargs: Keyword arguments to override the RUN_CONFIG.

    Returns:
      The current run.
    """
    run_config = getattr(self, 'RUN_ARGS', {})
    run_config.update(kwargs)
    if root_dir is not None:
      run_config['root_dir'] = root_dir
    if id is not None:
      run_config['id'] = id
    return self.run(**run_config)

  #
  # HTML views.
  #

  def output_link(
      self,
      run: Optional['Run'], relative_path: str
  ) -> str | None:
    """Returns the output path of the experiment."""
    if run is None:
      return None
    return self.link(run.output_path_for(self, relative_path))

  def _html_tree_view_summary_title(
      self,
      current_run: Optional['Run'] = None,
      interactive: bool = True,
  ):
    title, link, dir_link = self.id, None, None
    if current_run is not None:
      dir_link = self.output_link(current_run, '')
      if self.is_leaf:
        link = self.output_link(current_run, 'index.html')
      elif self.parent is None:
        title = str(current_run.id)
        link = self.output_link(current_run, 'summary.html')
    return pg.Html.element(
        'div',
        [
            # Experiment ID.
            pg.views.html.controls.Label(
                title,
                link=link,
                tooltip=pg.format(  # pytype: disable=wrong-arg-types
                    self,
                    verbose=False,
                    use_inferred=True,
                    hide_default_values=True,
                    exclude_keys=(
                        'root_dir', 'plugins', 'progress', 'usage_summary'
                    ),
                ),
                css_classes=['experiment-name'],
            ),
            # Experiment directory (if root or leaf).
            pg.views.html.controls.Label(  # pylint: disable=g-long-ternary
                '[dir]',
                link=dir_link,
                css_classes=['experiment-dir'],
            ) if dir_link is not None else None,
            # Progress bar.
            self.progress.to_html(
                extra_flags=dict(interactive=interactive),
            ),
            # Usage summary,
            self.usage_summary.to_html(
                extra_flags=dict(as_badge=True, interactive=interactive)
            ),
        ],
        css_classes=['experiment-summary']
    )

  def _html_tree_view_summary(
      self,
      *,
      view,
      name: str | None = None,
      extra_flags: dict[str, Any] | None = None,
      **kwargs
  ):
    extra_flags = extra_flags or {}
    if not extra_flags.get('card_view', True):
      return None

    kwargs.pop('title', None)
    kwargs.pop('enable_key_tooltip', None)
    kwargs.pop('enable_summary_tooltip', None)
    return view.summary(
        self,
        name=name if self.is_leaf else None,
        title=self._html_tree_view_summary_title(
            extra_flags.get('current_run', None),
            extra_flags.get('interactive', True)
        ),
        enable_key_tooltip=False,
        enable_summary_tooltip=False,
        **kwargs
    )

  def _html_tree_view_content(
      self,
      *,
      view,
      collapse_level: int | None = 1,
      extra_flags: dict[str, Any],
      **kwargs):
    return pg.Html.element(
        'div',
        [
            c.to_html(
                collapse_level=view.get_collapse_level(
                    (collapse_level, -1), 0
                ),
                name=f'#{i + 1}',
                extra_flags=extra_flags,
                **view.get_passthrough_kwargs(**kwargs)
            )
            for i, c in enumerate(self.children)
        ],
    )

  def _html_tree_view_css_styles(self) -> list[str]:
    return super()._html_tree_view_css_styles() + [
        """
        .experiment-summary {
            display: inline-block;
            font-weight: normal;
        }
        .experiment-name {
            font-weight: bold;
        }
        .experiment-dir.label {
            color: revert;
            margin-left: 0px;
            padding: 2px;
        }
        .usage-summary-badge {
          margin-left: 10px;
        }
        body {
          font: normal 16px "Roboto","Noto",sans-serif;
        }
        """
    ]


@pg.use_init_args(['children'])
class Suite(Experiment):
  """A suite of evaluations."""

  children: Annotated[
      list[Experiment], 'A list of child experiments.'
  ] = []

  @property
  def is_leaf(self) -> bool:
    """Returns whether the task is a leaf."""
    return False


class RunId(pg.Object):
  """Structured repreesentation a experiment run ID."""
  date: datetime.date
  number: int

  _REGEX = re.compile(r'^(\d{8})_(\d+)$')

  def dirname(self, root_dir: str | None = None) -> str:
    """Returns the directory name of the run ID."""
    dir_name = f'run_{self}'
    if root_dir is None:
      return dir_name
    return os.path.join(root_dir, dir_name)

  def __str__(self) -> str:
    """Returns the string representation of the run ID."""
    return f'{self.date.strftime("%Y%m%d")}_{self.number}'

  def __lt__(self, other: 'RunId') -> bool:
    """Returns whether the run ID is less than the other."""
    return self.date < other.date or (
        self.date == other.date and self.number < other.number
    )

  def _le__(self, other: 'RunId') -> bool:
    """Returns whether the run ID is less than or equal to the other."""
    return self == other or self < other

  def __gt__(self, other: 'RunId') -> bool:
    """Returns whether the run ID is greater than the other."""
    return other < self

  def __ge__(self, other: 'RunId') -> bool:
    """Returns whether the run ID is greater than or equal to the other."""
    return self == other or self > other

  def next(self) -> 'RunId':
    """Returns the next run ID."""
    return RunId(
        date=self.date,
        number=self.number + 1,
    )

  @classmethod
  def from_dirname(cls, dirname: str) -> Optional['RunId']:
    """Creates a run ID from the directory name."""
    if not dirname.startswith('run_'):
      return None
    run_id_str = dirname.removeprefix('run_')
    if cls.is_valid(run_id_str):
      return cls.from_id(run_id_str)
    return None

  @classmethod
  def is_valid(cls, run_id: str) -> bool:
    """Returns whether the run ID is valid."""
    return run_id in ('latest', 'new') or bool(cls._REGEX.match(run_id))

  @classmethod
  def from_id(
      cls,
      run_id: str | None,
      root_dir: str | None = None
  ) -> 'RunId':
    """Creates a run ID from the string representation."""
    if run_id is not None and not cls.is_valid(run_id):
      raise ValueError(
          f'`run_id` must be one of `latest`, `new` and a '
          f'datetime string in format `%Y%m%d%_<number>` (e.g. 20240101_1). '
          f'Encountered: {run_id!r}.'
      )
    if run_id in (None, 'latest', 'new'):
      if root_dir is None:
        raise ValueError(
            '`root_dir` must be provided for `latest` or `new` run ID.'
        )
      if run_id == 'latest':
        run_id = cls.get_latest(root_dir)
        if run_id is None:
          raise ValueError(
              f'There are no previous runs under the root directory: '
              f'{root_dir}. Consider running the experiment using `new` as id.'
          )
        return run_id
      if run_id == 'new':
        return cls.new(root_dir)
      return cls.get_latest(root_dir) or cls.new()

    assert run_id is not None
    date_str, number_str = run_id.split('_')
    return cls(
        date=datetime.datetime.strptime(date_str, '%Y%m%d').date(),
        number=int(number_str),
    )

  @classmethod
  def get_latest(cls, root_dir: str) -> Optional['RunId']:
    """Returns the latest run ID under the root directory."""
    if not pg.io.isdir(root_dir):
      return None
    run_ids = [
        RunId.from_dirname(dirname)
        for dirname in pg.io.listdir(root_dir)
    ]
    run_ids = [run_id for run_id in run_ids if run_id is not None]
    if not run_ids:
      return None
    return max(run_ids)

  @classmethod
  def new(cls, root_dir: str | None = None) -> 'RunId':
    """Creates a new run ID."""
    latest = None if root_dir is None else cls.get_latest(root_dir)
    if latest is not None and latest.date == datetime.date.today():
      return latest.next()
    return cls(
        date=datetime.date.today(),
        number=1,
    )


class Run(pg.Object, pg.views.html.HtmlTreeView.Extension):
  """A run of an experiment."""

  root_dir: Annotated[
      str,
      'The root of the output directory of the experiment.'
  ]

  id: Annotated[
      RunId,
      (
          'The ID of the current run.'
      )
  ]

  experiment: Annotated[
      Experiment,
      'The root experiment to run.'
  ]

  warm_start_from: Annotated[
      str | None,
      (
          'The directory for a previous run to warm start from.'
      )
  ] = None

  example_ids: Annotated[
      list[int] | None,
      (
          'The example IDs to run. If None, it will run all examples. '
          'Though '
      )
  ] = None

  shuffle_inputs: Annotated[
      bool,
      (
          'If True, the order of evaluating examples will be shuffled.'
          'Otherwise an increasing order will be used.'
      )
  ] = False

  raise_if_has_error: Annotated[
      bool,
      (
          'If True, it will raise an error if any example fails.'
      )
  ] = False

  reevaluate_upon_previous_errors: Annotated[
      bool,
      (
          'If True, it will reevaluate the example if the previously '
          'checkpointed run has error. Otherwise, it will skip the example.'
      )
  ] = True

  note: Annotated[
      str | None,
      'The user note for the current run.'
  ] = None

  tags: Annotated[
      list[str],
      'The user tags for the current run.'
  ] = []

  reprocess: Annotated[
      bool | list[int],
      (
          'If True, it will reprocess all examples under the current '
          'run directory. If a list of integers, examples of the given IDS '
          'will be reprocessed.'
      )
  ] = False

  generate_example_html: Annotated[
      Literal['new', 'all', 'no'] | list[int],
      (
          'If "new", generate HTML files for all newly processed examples, '
          'and keep/copy existing HTML files for unchanged examples. '
          'If "all", generate HTML files for all examples. '
          'If "no", do not generate HTML files for any examples. '
          'If a list of example IDs, generate HTML files for the specified '
          'examples.'
      )
  ] = 'new'

  filter: Annotated[
      Callable[[Experiment], bool] | None,
      'A filter to decide whether a leaf experiment should be run or not.'
  ] = None

  process_timeout: Annotated[
      int | None,
      'Timeout for each evaluation example.'
  ] = None

  use_cache: Annotated[
      Literal['global', 'per_dataset', 'no'],
      (
          'The cache policy for the runner. If `global`, the runner will use '
          'the cache for all evaluations. If `per_dataset`, the runner will '
          'use the cache for each evaluation. If `no`, the runner will not '
          'use the cache.'
      )
  ] = 'per_dataset'

  @property
  def output_root(self) -> str:
    """Returns the root directory of the experiment."""
    return self.id.dirname(self.root_dir)

  @property
  def input_root(self) -> str:
    """Returns the input root d."""
    return self.warm_start_from if self.warm_start_from else self.output_root

  def output_dir(self, experiment: Experiment) -> str:
    """Returns the output directory of the experiment."""
    if experiment.is_leaf:
      return os.path.join(self.output_root, experiment.id.replace('@', '/'))
    return self.output_root

  def output_link(self, relative_path: str):
    """Returns the output link for the relative path."""
    return Experiment.link(os.path.join(self.output_root, relative_path))

  def input_dir(self, experiment: Experiment) -> str:
    """Returns the input directory of the experiment."""
    if experiment.is_leaf:
      return os.path.join(self.input_root, experiment.id.replace('@', '/'))
    return self.input_root

  def input_path_for(self, experiment: Experiment, relative_path: str) -> str:
    """Returns the input path for the experiment."""
    return os.path.join(self.input_dir(experiment), relative_path)

  def output_path_for(self, experiment: Experiment, relative_path: str) -> str:
    """Returns the output path for the experiment."""
    return os.path.join(self.output_dir(experiment), relative_path)

  def examples_to_evaluate(self, experiment: Experiment) -> set[int]:
    """Returns the example IDs to evaluate."""
    if not experiment.is_leaf:
      return set()
    return set(
        self.example_ids if self.example_ids else
        range(1, experiment.num_examples + 1)
    )

  def examples_to_reprocess(self, experiment: Experiment) -> set[int]:
    """Returns the example IDs to reprocess per request."""
    if not self.reprocess:
      return set()
    reprocess_ids = self.examples_to_evaluate(experiment)
    if isinstance(self.reprocess, list):
      reprocess_ids &= set(self.reprocess)
    return reprocess_ids

  def examples_to_load(self, experiment: Experiment) -> set[int]:
    """Returns the example IDs to load from checkpoint files.."""
    load_ids = self.examples_to_evaluate(experiment)
    if isinstance(self.generate_example_html, list):
      load_ids |= set(self.generate_example_html)
    load_ids -= self.examples_to_reprocess(experiment)
    return load_ids

  def examples_to_load_metadata(self, experiment: Experiment) -> set[int]:
    """Returns the example IDs to load the metadata."""
    load_metadata_ids = set()
    if isinstance(self.generate_example_html, list):
      load_metadata_ids = set(self.generate_example_html)
    elif self.generate_example_html == 'all':
      load_metadata_ids = self.examples_to_evaluate(experiment)
    load_metadata_ids -= self.examples_to_reprocess(experiment)
    return load_metadata_ids


class Runner(pg.Object):
  """Interface for experiment runner."""

  # Class-level variable for registering the runner.
  NAME = None

  _REGISTRY = {}

  current_run: Annotated[
      Run,
      'The current run.'
  ]

  plugins: Annotated[
      list['Plugin'],
      'The plugins for the runner.'
  ] = []

  def __init_subclass__(cls):
    super().__init_subclass__()
    if inspect.isabstract(cls):
      return
    if cls.NAME is None:
      raise ValueError(
          'Runner class must define a NAME constant. '
          'Please use the same constant in the runner class.'
      )
    cls._REGISTRY[cls.NAME] = cls

  @abc.abstractmethod
  def run(self) -> None:
    """Runs a evaluation task."""

  @classmethod
  def create(cls, runner: str, **kwargs) -> 'Runner':
    """Creates a runner instance by ID and kwargs."""
    return cls._REGISTRY[runner](**kwargs)


class Plugin(lf.Component):
  """Base class for experiment plugins."""

  def on_run_start(
      self,
      runner: Runner,
      root: Experiment
  ) -> None:
    """Called when a runner is started."""

  def on_run_complete(
      self,
      runner: Runner,
      root: Experiment
  ) -> None:
    """Called when a runner is complete."""

  def on_run_abort(
      self,
      runner: Runner,
      root: Experiment,
      error: BaseException,
  ) -> None:
    """Called when a runner is aborted."""

  def on_experiment_start(
      self,
      runner: Runner,
      experiment: Experiment
  ) -> None:
    """Called when an evaluation is started."""

  def on_experiment_skipped(
      self,
      runner: Runner,
      experiment: Experiment
  ) -> None:
    """Called when an experiment (both leaf and non-leaf) is skipped."""

  def on_experiment_complete(
      self,
      runner: Runner,
      experiment: Experiment
  ) -> None:
    """Called when an experiment (both leaf and non-leaf) is complete."""

  def on_experiment_abort(
      self,
      runner: Runner,
      experiment: Experiment,
      error: BaseException,
  ) -> None:
    """Called when an experiment (both leaf and non-leaf) is aborted."""

  def on_example_start(
      self,
      runner: Runner,
      experiment: Experiment,
      example: example_lib.Example
  ) -> None:
    """Called when an example is about to be evaluated."""

  def on_example_complete(
      self,
      runner: Runner,
      experiment: Experiment,
      example: example_lib.Example
  ) -> None:
    """Called when an example is evaluated."""



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Language models from OpenAI."""

import datetime
import functools
import os
from typing import Annotated, Any, Final

import langfun.core as lf
from langfun.core.llms import openai_compatible
import pyglove as pg


class OpenAIModelInfo(lf.ModelInfo):
  """OpenAI model info."""

  # Constants for supported MIME types.
  INPUT_IMAGE_TYPES = [
      'image/png',
      'image/jpeg',
      'image/gif',
      'image/webp',
  ]

  LINKS = dict(
      models='https://platform.openai.com/docs/models',
      pricing='https://openai.com/api/pricing/',
      rate_limits='https://platform.openai.com/docs/guides/rate-limits',
      error_codes='https://platform.openai.com/docs/guides/error-codes',
  )

  provider: Final[str] = 'OpenAI'  # pylint: disable=invalid-name


#
# !!! Please sort models by model family and model_id (time descending).
#

SUPPORTED_MODELS = [
    # GPT-4.1 models
    OpenAIModelInfo(
        model_id='gpt-4.1',
        alias_for='gpt-4.1-2025-04-14',
        in_service=True,
        model_type='instruction-tuned',
        description='GPT 4.1 model (latest stable).',
        url='https://platform.openai.com/docs/models/gpt-4.1',
        input_modalities=OpenAIModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=1_047_576,
            max_output_tokens=32_768,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.50,
            cost_per_1m_input_tokens=2.0,
            cost_per_1m_output_tokens=8.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=30_000_000,
        ),
    ),
    # o3 models.
    OpenAIModelInfo(
        model_id='o3',
        alias_for='o3-2025-04-16',
        in_service=True,
        model_type='thinking',
        description='GPT O3 model.',
        url='https://platform.openai.com/docs/models/o3',
        input_modalities=OpenAIModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=200_000,
            max_output_tokens=100_000,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=2.5,
            cost_per_1m_input_tokens=10.0,
            cost_per_1m_output_tokens=40.0,
        ),
        # Set to 1/3 of Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=3_000,
            max_tokens_per_minute=10_000_000,
        ),
    ),
    # o4-mini models.
    OpenAIModelInfo(
        model_id='o4-mini',
        alias_for='o4-mini-2025-04-16',
        in_service=True,
        model_type='thinking',
        description='GPT O4-mini model.',
        url='https://platform.openai.com/docs/models/o4-mini',
        input_modalities=OpenAIModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=200_000,
            max_output_tokens=100_000,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.275,
            cost_per_1m_input_tokens=1.1,
            cost_per_1m_output_tokens=4.4,
        ),
        # Set to 1/3 of Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=50_000_000,
        ),
    ),
    # GPT-4.5 models
    OpenAIModelInfo(
        model_id='gpt-4.5-preview-2025-02-27',
        alias_for='gpt-4.5-preview-2025-02-27',
        in_service=True,
        model_type='instruction-tuned',
        description='GPT 4.5 preview model.',
        url='https://platform.openai.com/docs/models#gpt-4-5',
        input_modalities=OpenAIModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=16_384,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=37.5,
            cost_per_1m_input_tokens=75,
            cost_per_1m_output_tokens=150.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=2_000_000,
        ),
    ),
    # o3-mini models.
    OpenAIModelInfo(
        model_id='o3-mini',
        alias_for='o3-mini-2025-01-31',
        in_service=True,
        model_type='thinking',
        description='GPT O3-mini model (latest).',
        url='https://platform.openai.com/docs/models#o3-mini',
        input_modalities=OpenAIModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=200_000,
            max_output_tokens=100_000,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.55,
            cost_per_1m_input_tokens=1.1,
            cost_per_1m_output_tokens=4.4,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=10_000_000,
        ),
    ),
    OpenAIModelInfo(
        model_id='o3-mini-2025-01-31',
        in_service=True,
        model_type='thinking',
        description='GPT O3-mini model (01/31/2025).',
        url='https://platform.openai.com/docs/models#o3-mini',
        release_date=datetime.datetime(2025, 1, 31),
        input_modalities=OpenAIModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=200_000,
            max_output_tokens=100_000,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.55,
            cost_per_1m_input_tokens=1.1,
            cost_per_1m_output_tokens=4.4,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=10_000_000,
        ),
    ),
    # o1-mini models.
    OpenAIModelInfo(
        model_id='o1-mini',
        alias_for='o1-mini-2024-09-12',
        in_service=True,
        model_type='thinking',
        description='GPT O1-mini model (latest).',
        url='https://platform.openai.com/docs/models#o1',
        input_modalities=OpenAIModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=65_536,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.55,
            cost_per_1m_input_tokens=1.1,
            cost_per_1m_output_tokens=4.4,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=10_000_000,
        ),
    ),
    OpenAIModelInfo(
        model_id='o1-mini-2024-09-12',
        in_service=True,
        model_type='thinking',
        description='GPT O1-mini model (09/12/2024).',
        url='https://platform.openai.com/docs/models#o1',
        input_modalities=OpenAIModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=65_536,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.55,
            cost_per_1m_input_tokens=1.1,
            cost_per_1m_output_tokens=4.4,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=10_000_000,
        ),
    ),
    OpenAIModelInfo(
        model_id='o1-preview',
        alias_for='o1-preview-2024-09-12',
        in_service=True,
        model_type='thinking',
        description='GPT O1-preview model (latest).',
        url='https://platform.openai.com/docs/models#o1',
        input_modalities=OpenAIModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=32_768,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=7.5,
            cost_per_1m_input_tokens=15.0,
            cost_per_1m_output_tokens=60.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=10_000_000,
        ),
    ),
    OpenAIModelInfo(
        model_id='o1-preview-2024-09-12',
        in_service=True,
        model_type='thinking',
        description='GPT O1-preview model (09/12/2024).',
        url='https://platform.openai.com/docs/models#o1',
        release_date=datetime.datetime(2024, 9, 12),
        input_modalities=OpenAIModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=32_768,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=7.5,
            cost_per_1m_input_tokens=15.0,
            cost_per_1m_output_tokens=60.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=2_000_000,
        ),
    ),
    # o1 models.
    OpenAIModelInfo(
        model_id='o1',
        alias_for='o1-2024-12-17',
        in_service=True,
        model_type='thinking',
        description='GPT O1 model (latest).',
        url='https://platform.openai.com/docs/models#o1',
        input_modalities=OpenAIModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=200_000,
            max_output_tokens=100_000,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=7.5,
            cost_per_1m_input_tokens=15.0,
            cost_per_1m_output_tokens=60.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=2_000_000,
        ),
    ),
    OpenAIModelInfo(
        model_id='o1-2024-12-17',
        in_service=True,
        model_type='thinking',
        description='GPT O1 model (12/17/2024).',
        url='https://platform.openai.com/docs/models#o1',
        release_date=datetime.datetime(2024, 12, 17),
        input_modalities=OpenAIModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=200_000,
            max_output_tokens=100_000,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=7.5,
            cost_per_1m_input_tokens=15.0,
            cost_per_1m_output_tokens=60.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=2_000_000,
        ),
    ),
    # GPT-4o-mini models
    OpenAIModelInfo(
        model_id='gpt-4o-mini',
        alias_for='gpt-4o-mini-2024-07-18',
        in_service=True,
        model_type='instruction-tuned',
        description='GPT 4o mini model (latest).',
        url='https://platform.openai.com/docs/models#gpt-4o-mini',
        input_modalities=OpenAIModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=16_384,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.075,
            cost_per_1m_input_tokens=0.15,
            cost_per_1m_output_tokens=0.6,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=10_000_000,
        ),
    ),
    OpenAIModelInfo(
        model_id='gpt-4o-mini-2024-07-18',
        in_service=True,
        model_type='instruction-tuned',
        description='GPT 4o mini model (07/18/2024).',
        url='https://platform.openai.com/docs/models#gpt-4o-mini',
        release_date=datetime.datetime(2024, 7, 18),
        input_modalities=OpenAIModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=16_384,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.075,
            cost_per_1m_input_tokens=0.15,
            cost_per_1m_output_tokens=0.6,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=10_000_000,
        ),
    ),
    # GPT-4o models
    OpenAIModelInfo(
        model_id='gpt-4o',
        alias_for='gpt-4o-2024-08-06',
        in_service=True,
        model_type='instruction-tuned',
        description='GPT 4o model (latest stable).',
        url='https://platform.openai.com/docs/models#gpt-4o',
        input_modalities=OpenAIModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=16_384,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=1.25,
            cost_per_1m_input_tokens=2.5,
            cost_per_1m_output_tokens=10.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=2_000_000,
        ),
    ),
    OpenAIModelInfo(
        model_id='gpt-4o-2024-11-20',
        in_service=True,
        model_type='instruction-tuned',
        description='GPT 4o model (11/20/2024).',
        url='https://platform.openai.com/docs/models#gpt-4o',
        release_date=datetime.datetime(2024, 11, 20),
        input_modalities=OpenAIModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=16_384,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=1.25,
            cost_per_1m_input_tokens=2.5,
            cost_per_1m_output_tokens=10.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=2_000_000,
        ),
    ),
    OpenAIModelInfo(
        model_id='gpt-4o-2024-08-06',
        in_service=True,
        model_type='instruction-tuned',
        description='GPT 4o model (08/06/2024).',
        url='https://platform.openai.com/docs/models#gpt-4o',
        input_modalities=OpenAIModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=16_384,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=1.25,
            cost_per_1m_input_tokens=2.5,
            cost_per_1m_output_tokens=10.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=2_000_000,
        ),
    ),
    OpenAIModelInfo(
        model_id='gpt-4o-2024-05-13',
        in_service=True,
        model_type='instruction-tuned',
        description='GPT 4o model (05/13/2024).',
        url='https://platform.opedsnai.com/docs/models#gpt-4o',
        release_date=datetime.datetime(2024, 5, 13),
        input_modalities=OpenAIModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=16_384,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=None,
            cost_per_1m_input_tokens=5.0,
            cost_per_1m_output_tokens=15.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=2_000_000,
        ),
    ),
    OpenAIModelInfo(
        model_id='chatgpt-4o-latest',
        in_service=True,
        model_type='instruction-tuned',
        description='GPT 4o model ChatGPT version (latest).',
        url='https://platform.openai.com/docs/models#gpt-4o',
        input_modalities=OpenAIModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=16_384,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=None,
            cost_per_1m_input_tokens=5.0,
            cost_per_1m_output_tokens=15.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=2_000_000,
        ),
    ),
    # GPT-4 Turbo models.
    OpenAIModelInfo(
        model_id='gpt-4-turbo',
        alias_for='gpt-4-turbo-2024-04-09',
        in_service=True,
        model_type='instruction-tuned',
        description='GPT 4 Turbo model (latest).',
        url='https://platform.openai.com/docs/models#gpt-4-turbo-and-gpt-4',
        input_modalities=OpenAIModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=4_096,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=None,
            cost_per_1m_input_tokens=10.0,
            cost_per_1m_output_tokens=30.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=800_000,
        ),
    ),
    OpenAIModelInfo(
        model_id='gpt-4-turbo-2024-04-09',
        in_service=True,
        model_type='instruction-tuned',
        description='GPT 4 Turbo model (04/09/2024).',
        url='https://platform.openai.com/docs/models#gpt-4-turbo-and-gpt-4',
        release_date=datetime.datetime(2024, 4, 9),
        input_modalities=OpenAIModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=4_096,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=None,
            cost_per_1m_input_tokens=10.0,
            cost_per_1m_output_tokens=30.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=800_000,
        ),
    ),
    OpenAIModelInfo(
        model_id='gpt-4-turbo-preview',
        alias_for='gpt-4-0125-preview',
        in_service=True,
        model_type='instruction-tuned',
        description='GPT 4 Turbo preview model (latest).',
        url='https://platform.openai.com/docs/models#gpt-4-turbo-and-gpt-4',
        input_modalities=OpenAIModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=4_096,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=None,
            cost_per_1m_input_tokens=10.0,
            cost_per_1m_output_tokens=30.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=800_000,
        ),
    ),
    OpenAIModelInfo(
        model_id='gpt-4-0125-preview',
        in_service=True,
        model_type='instruction-tuned',
        description='GPT 4 Turbo preview model (01/25/2024).',
        url='https://platform.openai.com/docs/models#gpt-4-turbo-and-gpt-4',
        release_date=datetime.datetime(2024, 1, 25),
        input_modalities=OpenAIModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=4_096,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=None,
            cost_per_1m_input_tokens=10.0,
            cost_per_1m_output_tokens=30.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=800_000,
        ),
    ),
    OpenAIModelInfo(
        model_id='gpt-4-1106-preview',
        in_service=True,
        model_type='instruction-tuned',
        description='GPT 4 Turbo preview model (11/06/2024).',
        url='https://platform.openai.com/docs/models#gpt-4-turbo-and-gpt-4',
        release_date=datetime.datetime(2024, 11, 6),
        input_modalities=OpenAIModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=4_096,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=None,
            cost_per_1m_input_tokens=10.0,
            cost_per_1m_output_tokens=30.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=800_000,
        ),
    ),
    # GPT-4 models.
    OpenAIModelInfo(
        model_id='gpt-4',
        alias_for='gpt-4-0613',
        in_service=True,
        model_type='instruction-tuned',
        description='GPT 4 model (latest).',
        url='https://platform.openai.com/docs/models#gpt-4-turbo-and-gpt-4',
        input_modalities=lf.ModelInfo.TEXT_INPUT_ONLY,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=8_192,
            max_output_tokens=8_192,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=None,
            cost_per_1m_input_tokens=30.0,
            cost_per_1m_output_tokens=60.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=300_000,
        ),
    ),
    OpenAIModelInfo(
        model_id='gpt-4-0613',
        in_service=True,
        model_type='instruction-tuned',
        description='GPT 4 model (06/13/2023).',
        url='https://platform.openai.com/docs/models#gpt-4-turbo-and-gpt-4',
        release_date=datetime.datetime(2023, 6, 13),
        input_modalities=lf.ModelInfo.TEXT_INPUT_ONLY,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=8_192,
            max_output_tokens=8_192,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=None,
            cost_per_1m_input_tokens=30.0,
            cost_per_1m_output_tokens=60.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=300_000,
        ),
    ),
    OpenAIModelInfo(
        model_id='gpt-4-0314',
        in_service=False,
        model_type='instruction-tuned',
        description='GPT 4 model (03/14/2023).',
        url='https://platform.openai.com/docs/models#gpt-4-turbo-and-gpt-4',
        release_date=datetime.datetime(2023, 3, 14),
        input_modalities=lf.ModelInfo.TEXT_INPUT_ONLY,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=8_192,
            max_output_tokens=8_192,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=None,
            cost_per_1m_input_tokens=30.0,
            cost_per_1m_output_tokens=60.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=300_000,
        ),
    ),
    # GPT-4 32K models.
    OpenAIModelInfo(
        model_id='gpt-4-32k',
        alias_for='gpt-4-32k-0613',
        in_service=False,
        model_type='instruction-tuned',
        description='GPT 4 32K model (latest).',
        url='https://platform.openai.com/docs/models#gpt-4-turbo-and-gpt-4',
        input_modalities=lf.ModelInfo.TEXT_INPUT_ONLY,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=32_768,
            max_output_tokens=8_192,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=None,
            cost_per_1m_input_tokens=60.0,
            cost_per_1m_output_tokens=120.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=300_000,
        ),
    ),
    OpenAIModelInfo(
        model_id='gpt-4-32k-0613',
        in_service=False,
        model_type='instruction-tuned',
        description='GPT 4 32K model (06/13/2023).',
        url='https://platform.openai.com/docs/models#gpt-4-turbo-and-gpt-4',
        release_date=datetime.datetime(2023, 6, 13),
        input_modalities=lf.ModelInfo.TEXT_INPUT_ONLY,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=32_768,
            max_output_tokens=8_192,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=None,
            cost_per_1m_input_tokens=60.0,
            cost_per_1m_output_tokens=120.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=300_000,
        ),
    ),
    OpenAIModelInfo(
        model_id='gpt-4-32k-0314',
        in_service=False,
        model_type='instruction-tuned',
        description='GPT 4 32K model (03/14/2023).',
        url='https://platform.openai.com/docs/models#gpt-4-turbo-and-gpt-4',
        release_date=datetime.datetime(2023, 3, 14),
        input_modalities=lf.ModelInfo.TEXT_INPUT_ONLY,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=32_768,
            max_output_tokens=8_192,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=None,
            cost_per_1m_input_tokens=60.0,
            cost_per_1m_output_tokens=120.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=300_000,
        ),
    ),
    # GPT 3.5 Turbo models.
    OpenAIModelInfo(
        model_id='gpt-3.5-turbo',
        alias_for='gpt-3.5-turbo-0125',
        in_service=True,
        model_type='instruction-tuned',
        description='GPT 3.5 Turbo model (latest).',
        url='https://platform.openai.com/docs/models#gpt-3-5-turbo',
        input_modalities=lf.ModelInfo.TEXT_INPUT_ONLY,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=16_384,
            max_output_tokens=4_096,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=None,
            cost_per_1m_input_tokens=0.5,
            cost_per_1m_output_tokens=1.5,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=10_000_000,
        ),
    ),
    OpenAIModelInfo(
        model_id='gpt-3.5-turbo-0125',
        in_service=True,
        release_date=datetime.datetime(2024, 1, 25),
        model_type='instruction-tuned',
        description='GPT 3.5 Turbo model (01/25/2024).',
        url='https://platform.openai.com/docs/models#gpt-3-5-turbo',
        input_modalities=lf.ModelInfo.TEXT_INPUT_ONLY,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=16_384,
            max_output_tokens=4_096,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=None,
            cost_per_1m_input_tokens=0.5,
            cost_per_1m_output_tokens=1.5,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=10_000_000,
        ),
    ),
    OpenAIModelInfo(
        model_id='gpt-3.5-turbo-1106',
        in_service=True,
        release_date=datetime.datetime(2023, 11, 6),
        model_type='instruction-tuned',
        description='GPT 3.5 Turbo model (11/06/2023).',
        url='https://platform.openai.com/docs/models#gpt-3-5-turbo',
        input_modalities=lf.ModelInfo.TEXT_INPUT_ONLY,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=16_384,
            max_output_tokens=4_096,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=None,
            cost_per_1m_input_tokens=1.0,
            cost_per_1m_output_tokens=2.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=10_000_000,
        ),
    ),
    OpenAIModelInfo(
        model_id='gpt-3.5-turbo-0613',
        in_service=False,
        release_date=datetime.datetime(2023, 6, 13),
        model_type='instruction-tuned',
        description='GPT 3.5 Turbo model (06/13/2023).',
        url='https://platform.openai.com/docs/models#gpt-3-5-turbo',
        input_modalities=lf.ModelInfo.TEXT_INPUT_ONLY,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=16_384,
            max_output_tokens=4_096,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=None,
            cost_per_1m_input_tokens=1.5,
            cost_per_1m_output_tokens=2.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=10_000_000,
        ),
    ),
    # GPT 3.5 Turbo 16K models.
    OpenAIModelInfo(
        model_id='gpt-3.5-turbo-16k',
        alias_for='gpt-3.5-turbo-16k-0613',
        in_service=True,
        model_type='instruction-tuned',
        description='GPT 3.5 Turbo 16K model (latest).',
        url='https://platform.openai.com/docs/models#gpt-3-5-turbo',
        input_modalities=lf.ModelInfo.TEXT_INPUT_ONLY,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=16_385,
            max_output_tokens=4_096,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=None,
            cost_per_1m_input_tokens=3.0,
            cost_per_1m_output_tokens=4.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=10_000_000,
        ),
    ),
    OpenAIModelInfo(
        model_id='gpt-3.5-turbo-16k-0613',
        in_service=False,
        model_type='instruction-tuned',
        description='GPT 3.5 Turbo 16K model (06/13/2023).',
        url='https://platform.openai.com/docs/models#gpt-3-5-turbo',
        release_date=datetime.datetime(2023, 6, 13),
        input_modalities=lf.ModelInfo.TEXT_INPUT_ONLY,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=16_385,
            max_output_tokens=4_096,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=None,
            cost_per_1m_input_tokens=3.0,
            cost_per_1m_output_tokens=4.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=10_000_000,
        ),
    ),
    OpenAIModelInfo(
        model_id='gpt-3.5-turbo-16k-0301',
        in_service=False,
        model_type='instruction-tuned',
        description='GPT 3.5 Turbo 16K model (03/01/2023).',
        url='https://platform.openai.com/docs/models#gpt-3-5-turbo',
        release_date=datetime.datetime(2023, 3, 1),
        input_modalities=lf.ModelInfo.TEXT_INPUT_ONLY,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=16_385,
            max_output_tokens=4_096,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=None,
            cost_per_1m_input_tokens=3.0,
            cost_per_1m_output_tokens=4.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=10_000,
            max_tokens_per_minute=10_000_000,
        ),
    ),
    # GPT 3.5 models.
    OpenAIModelInfo(
        model_id='text-davinci-003',
        in_service=False,
        model_type='instruction-tuned',
        description='ChatGPT 3.5 model.',
        input_modalities=lf.ModelInfo.TEXT_INPUT_ONLY,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=16_384,
            max_output_tokens=4_096,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=None,
            cost_per_1m_input_tokens=3.0,
            cost_per_1m_output_tokens=3.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=None,
            max_tokens_per_minute=None,
        ),
    ),
    # GPT 3 models.
    OpenAIModelInfo(
        model_id='babbage-002',
        in_service=True,
        model_type='pretrained',
        description='GPT3 base model babagge-002',
        url='https://platform.openai.com/docs/models#gpt-base',
        input_modalities=lf.ModelInfo.TEXT_INPUT_ONLY,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=16_384,
            max_output_tokens=4_096,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=None,
            cost_per_1m_input_tokens=2.0,
            cost_per_1m_output_tokens=2.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=3_000,
            max_tokens_per_minute=250_000,
        ),
    ),
    OpenAIModelInfo(
        model_id='davinci-002',
        in_service=True,
        model_type='pretrained',
        description='GPT3 base model Davinci-002 ',
        url='https://platform.openai.com/docs/models#gpt-base',
        input_modalities=lf.ModelInfo.TEXT_INPUT_ONLY,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=16_384,
            max_output_tokens=4_096,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=None,
            cost_per_1m_input_tokens=2.0,
            cost_per_1m_output_tokens=2.0,
        ),
        # Tier 5 rate limits.
        rate_limits=lf.ModelInfo.RateLimits(
            max_requests_per_minute=3_000,
            max_tokens_per_minute=250_000,
        ),
    ),
]

_SUPPORTED_MODELS_BY_MODEL_ID = {m.model_id: m for m in SUPPORTED_MODELS}


@lf.use_init_args(['model'])
class OpenAI(openai_compatible.OpenAICompatible):
  """OpenAI model."""

  model: pg.typing.Annotated[
      pg.typing.Enum(
          pg.MISSING_VALUE, list(s.model_id for s in SUPPORTED_MODELS)
      ),
      'The name of the model to use.',
  ]

  api_endpoint: str = 'https://api.openai.com/v1/chat/completions'

  api_key: Annotated[
      str | None,
      (
          'API key. If None, the key will be read from environment variable '
          "'OPENAI_API_KEY'."
      ),
  ] = None

  organization: Annotated[
      str | None,
      (
          'Organization. If None, the key will be read from environment '
          "variable 'OPENAI_ORGANIZATION'. Based on the value, usages from "
          "these API requests will count against the organization's quota. "
      ),
  ] = None

  project: Annotated[
      str | None,
      (
          'Project. If None, the key will be read from environment '
          "variable 'OPENAI_PROJECT'. Based on the value, usages from "
          "these API requests will count against the project's quota. "
      ),
  ] = None

  def _on_bound(self):
    super()._on_bound()
    self._api_key = None
    self._organization = None
    self._project = None
    self.__dict__.pop('model_info', None)

  def _initialize(self):
    api_key = self.api_key or os.environ.get('OPENAI_API_KEY', None)
    if not api_key:
      raise ValueError(
          'Please specify `api_key` during `__init__` or set environment '
          'variable `OPENAI_API_KEY` with your OpenAI API key.'
      )
    self._api_key = api_key
    self._organization = self.organization or os.environ.get(
        'OPENAI_ORGANIZATION', None
    )
    self._project = self.project or os.environ.get('OPENAI_PROJECT', None)

  @property
  def headers(self) -> dict[str, Any]:
    assert self._api_initialized
    headers = super().headers
    headers['Authorization'] = f'Bearer {self._api_key}'
    if self._organization:
      headers['OpenAI-Organization'] = self._organization
    if self._project:
      headers['OpenAI-Project'] = self._project
    return headers

  @functools.cached_property
  def model_info(self) -> OpenAIModelInfo:
    return _SUPPORTED_MODELS_BY_MODEL_ID[self.model]

  @classmethod
  def dir(cls):
    return [s.model_id for s in SUPPORTED_MODELS if s.in_service]

  def _request_args(
      self, options: lf.LMSamplingOptions) -> dict[str, Any]:
    # Reasoning models (o1 series) does not support `logprobs` by 2024/09/12.
    if options.logprobs and self.model.startswith(('o1-', 'o3-')):
      raise RuntimeError('`logprobs` is not supported on {self.model!r}.')
    return super()._request_args(options)


class Gpt41(OpenAI):
  """GPT-4.1."""
  model = 'gpt-4.1'


class GptO4Mini(OpenAI):
  """GPT O4 mini."""

  model = 'o4-mini'
  timeout = None


class GptO3(OpenAI):
  """GPT O3."""

  model = 'o3'
  timeout = None


class Gpt45Preview_20250227(OpenAI):  # pylint: disable=invalid-name
  """Gpt-4.5 Preview 2025-02-27."""

  model = 'gpt-4.5-preview-2025-02-27'


class GptO3Mini(OpenAI):
  """GPT-O3-mini."""
  model = 'o3-mini'
  timeout = None


class GptO1(OpenAI):
  """GPT-O1."""
  model = 'o1'
  timeout = None


class GptO1Preview(OpenAI):
  """GPT-O1."""
  model = 'o1-preview'
  timeout = None


class GptO1Preview_20240912(OpenAI):   # pylint: disable=invalid-name
  """GPT O1."""
  model = 'o1-preview-2024-09-12'
  timeout = None


class GptO1Mini(OpenAI):
  """GPT O1-mini."""
  model = 'o1-mini'
  timeout = None


class GptO1Mini_20240912(OpenAI):   # pylint: disable=invalid-name
  """GPT O1-mini."""
  model = 'o1-mini-2024-09-12'
  timeout = None


class Gpt4oMini(OpenAI):
  """GPT-4o Mini."""
  model = 'gpt-4o-mini'


class Gpt4oMini_20240718(OpenAI):  # pylint:disable=invalid-name
  """GPT-4o Mini."""
  model = 'gpt-4o-mini-2024-07-18'


class Gpt4o(OpenAI):
  """GPT-4o."""
  model = 'gpt-4o'


class Gpt4o_20241120(OpenAI):     # pylint:disable=invalid-name
  """GPT-4o version 2024-11-20."""
  model = 'gpt-4o-2024-11-20'


class Gpt4o_20240806(OpenAI):     # pylint:disable=invalid-name
  """GPT-4o version 2024-08-06."""
  model = 'gpt-4o-2024-08-06'


class Gpt4o_20240513(OpenAI):     # pylint:disable=invalid-name
  """GPT-4o version 2024-05-13."""
  model = 'gpt-4o-2024-05-13'


class Gpt4(OpenAI):
  """GPT-4."""
  model = 'gpt-4'


class Gpt4Turbo(Gpt4):
  """GPT-4 Turbo with 128K context window. Knowledge up to Dec. 2023."""
  model = 'gpt-4-turbo'


class Gpt4Turbo_20240409(Gpt4Turbo):  # pylint:disable=invalid-name
  """GPT-4 Turbo with 128K context window. Knowledge up to Dec. 2023."""
  model = 'gpt-4-turbo-2024-04-09'


class Gpt4TurboPreview(Gpt4):
  """GPT-4 Turbo Preview with 128k context window. Knowledge up to Dec. 2023."""
  model = 'gpt-4-turbo-preview'


class Gpt4TurboPreview_20240125(Gpt4TurboPreview):  # pylint: disable=invalid-name
  """GPT-4 Turbo Preview with 128k context window. Knowledge up to Dec. 2023."""
  model = 'gpt-4-0125-preview'


class Gpt4TurboPreview_20231106(Gpt4TurboPreview):  # pylint: disable=invalid-name
  """GPT-4 Turbo Preview with 128k context window. Knowledge up to Apr. 2023."""
  model = 'gpt-4-1106-preview'


class Gpt4_20230613(Gpt4):    # pylint:disable=invalid-name
  """GPT-4 @20230613. 8K context window. Knowledge up to 9-2021."""
  model = 'gpt-4-0613'


class Gpt4_32K(Gpt4):       # pylint:disable=invalid-name
  """Latest GPT-4 with 32K context window."""
  model = 'gpt-4-32k'


class Gpt4_32K_20230613(Gpt4_32K):    # pylint:disable=invalid-name
  """GPT-4 @20230613. 32K context window. Knowledge up to 9-2021."""
  model = 'gpt-4-32k-0613'


class Gpt35(OpenAI):
  """GPT-3.5. 4K max tokens, trained up on data up to Sep, 2021."""
  model = 'text-davinci-003'


class Gpt35Turbo(Gpt35):
  """Most capable GPT-3.5 model, 10x cheaper than GPT35 (text-davinci-003)."""
  model = 'gpt-3.5-turbo'


class Gpt35Turbo_20240125(Gpt35Turbo):   # pylint:disable=invalid-name
  """GPT-3.5 Turbo @20240125. 16K context window. Knowledge up to 09/2021."""
  model = 'gpt-3.5-turbo-0125'


class Gpt35Turbo_20231106(Gpt35Turbo):   # pylint:disable=invalid-name
  """Gpt3.5 Turbo @20231106. 16K context window. Knowledge up to 09/2021."""
  model = 'gpt-3.5-turbo-1106'


class Gpt35Turbo_20230613(Gpt35Turbo):   # pylint:disable=invalid-name
  """Gpt3.5 Turbo snapshot at 2023/06/13, with 4K context window size."""
  model = 'gpt-3.5-turbo-0613'


class Gpt35Turbo16K(Gpt35Turbo):
  """Latest GPT-3.5 model with 16K context window size."""
  model = 'gpt-3.5-turbo-16k'


class Gpt35Turbo16K_20230613(Gpt35Turbo):   # pylint:disable=invalid-name
  """Gtp 3.5 Turbo 16K 0613."""
  model = 'gpt-3.5-turbo-16k-0613'


def _register_openai_models():
  """Registers OpenAI models."""
  for m in SUPPORTED_MODELS:
    lf.LanguageModel.register(m.model_id, OpenAI)

_register_openai_models()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Scoring the output objects based on their inputs."""

from typing import Any, Type, Union

import langfun.core as lf
from langfun.core.structured import mapping
from langfun.core.structured import querying
from langfun.core.structured import schema as schema_lib
import pyglove as pg


def score(
    prompt: Union[str, pg.Symbolic] | list[str | pg.Symbolic],
    completions: list[str | pg.Symbolic],
    schema: Union[
        schema_lib.Schema, Type[Any], list[Type[Any]], dict[str, Any], None
    ] = None,
    *,
    lm: lf.LanguageModel | None = None,
    examples: list[mapping.MappingExample] | None = None,
    protocol: schema_lib.SchemaProtocol = 'python',
    return_scoring_results: bool = False,
    **kwargs,
) -> list[float] | list[lf.LMScoringResult]:
  """Scores the outputs based on the prompt.

  Examples:
    ```
    # Example 1: Scoring text output based on the user prompt.
    scores = lf.score('{{x}} + {{y}} =', ['1', '2', '3'], lm=lm, x=1, y=2)
    assert len(scores) == 3

    # Example 2: Scoring int output based on the formulated OOP prompt.
    scores = lf.score('1 + 1 =', [1, 2, 3], lm=lm)
    assert len(scores) == 3

    class Answer(pg.Object):
      result: int

    # Example 3: Scoring object output based on the formulated OOP prompt.
    scores = lf.score('1 + 1 =', [Answer(1), Answer(2), Answer(3)], lm=lm)
    assert len(scores) == 3

    # Example 4: Scoring object field value based on the formulated OOP prompt
    # and the generated tokens before the first `pg.oneof`.
    scores = lf.score('1 + 1 =', [Answer(pg.oneof([1, 2, 3]))], lm=lm)
    assert len(scores) == 3

    # Example 5: Scoring multiple prompt/completion pairs.
    scores = lf.score(
        ['1 + 1=', '2 + 3='],
        ['2', '4'],
        lm=lm
    )
    assert len(scores) == 2
    ```

  Args:
    prompt: The prompt(s) based on which each completion will be scored.
    completions: A list of strings or symbolic objects as the output.
    schema: The schema as the output type. If None, it will be inferred from
      the completions.
    lm: The language model used for scoring.
    examples: Fewshot exemplars used together with the prompt in getting the
      completions.
    protocol: The protocol for formulating the prompt based on objects.
    return_scoring_results: If True, returns a list of `lf.LMScoringResult`,
      otherwise returns a list of floats as the scores of each completion.
    **kwargs: Keyword arguments that are referred by the prompt.

  Returns:
    A list of floats or `lf.LMScoringResult` as the score of each completion.
  """
  if not completions:
    raise ValueError('`completions` must not be empty.')

  if schema is None:
    for c in completions:
      if schema is None:
        schema = type(c)
      elif schema is not type(c):
        raise ValueError(
            '`schema` cannot be inferred from completions of different types: '
            f'{[type(c) for c in completions]}.'
        )

  if isinstance(prompt, list):
    prompts = []
    for p in prompt:
      prompts.append(
          querying.query_prompt(
              p,
              schema,
              examples=examples,
              protocol=protocol,
              **kwargs,
          )
      )
    input_message = prompts
  else:
    input_message = querying.query_prompt(
        prompt,
        schema,
        examples=examples,
        protocol=protocol,
        **kwargs,
    )
  if lm is None:
    lm_override = lf.get_contextual_override('lm')
    if lm_override is None:
      raise ValueError('`lm` must be specified or provided from `lf.context`.')
    lm = lm_override.value

  completion_reprs = []
  for c in completions:
    if isinstance(c, mapping.MappingError):
      completion_reprs.append(c.lm_response)
    else:
      rep = mapping.MappingExample.value_repr(
          c, protocol=protocol, compact=False, verbose=False
      )

      # NOTE(daiyip): supporting scenario of scoring object field with
      # `pg.oneof`.
      oneof_pos = rep.find('OneOf(')
      if oneof_pos == -1:
        completion_reprs.append(rep)
      else:
        assert protocol == 'python', protocol
        if isinstance(input_message, list):
          raise ValueError(
              'Scoring on object fields using `pg.oneof` must share the '
              f'same prompt. Encountered: {prompt}'
          )
        input_message.text += '\n' + rep[:oneof_pos]
        oneof = _get_first_oneof(c)
        for v in oneof.candidates:
          completion_reprs.append(
              pg.format(
                  v,
                  python_format=True,
                  compact=False,
                  verbose=False,
                  root_indent=oneof.sym_path.depth
              )
          )

  results = lm.score(
      input_message,
      completion_reprs,
  )
  if return_scoring_results:
    return results
  return [r.score for r in results]


def _get_first_oneof(value: Any) -> pg.hyper.OneOf:
  """Gets the first pg.oneof from a symbolic object."""
  oneofs = []
  def select_oneofs(k, v, p):
    del k, p
    if isinstance(v, pg.hyper.OneOf):
      oneofs.append(v)
      return pg.TraverseAction.CONTINUE
    return pg.TraverseAction.ENTER
  pg.traverse(value, select_oneofs)
  assert oneofs
  return oneofs[0]


async def ascore(
    prompt: Union[str, pg.Symbolic] | list[str | pg.Symbolic],
    completions: list[str | pg.Symbolic],
    schema: Union[
        schema_lib.Schema, Type[Any], list[Type[Any]], dict[str, Any], None
    ] = None,
    *,
    lm: lf.LanguageModel | None = None,
    examples: list[mapping.MappingExample] | None = None,
    protocol: schema_lib.SchemaProtocol = 'python',
    return_scoring_results: bool = False,
    **kwargs,
) -> list[float] | list[lf.LMScoringResult]:
  """Async version of `lf.score`."""
  # TODO(daiyip): implement native async scoring.
  return await lf.invoke_async(
      score,
      prompt,
      completions,
      schema=schema,
      lm=lm,
      examples=examples,
      protocol=protocol,
      return_scoring_results=return_scoring_results,
      **kwargs,
  )



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Structures for location objects in an image."""

import math
import random
from typing import Optional, Union
import pyglove as pg


class Coordinate(pg.Object):
  """A coordinate in a 2D image."""
  x: int
  y: int

  def as_tuple(self) -> tuple[int, int]:
    """Returns the coordinate as a tuple."""
    return self.x, self.y

  def __add__(
      self, other: Union['Coordinate', tuple[int, int]]
  ) -> 'Coordinate':
    """Returns the coordinate plus the other coordinate."""
    other = Coordinate.from_value(other)
    return Coordinate(self.x + other.x, self.y + other.y)

  def __radd__(
      self, other: Union['Coordinate', tuple[int, int]]
  ) -> 'Coordinate':
    """Returns the coordinate plus the other coordinate."""
    return self + other

  def __sub__(
      self, other: Union['Coordinate', tuple[int, int]]
  ) -> 'Coordinate':
    """Returns the coordinate minus the other coordinate."""
    other = Coordinate.from_value(other)
    return Coordinate(self.x - other.x, self.y - other.y)

  def __rsub__(
      self, other: Union['Coordinate', tuple[int, int]]
  ) -> 'Coordinate':
    """Returns the coordinate minus the other coordinate."""
    other = Coordinate.from_value(other)
    return other - self

  def __mul__(self, ratio: float) -> 'Coordinate':
    """Returns the coordinate multiplied by a ratio."""
    return Coordinate(self.x * ratio, self.y * ratio)

  def __rmul__(self, ratio: float) -> 'Coordinate':
    """Returns the coordinate multiplied by a ratio."""
    return self * ratio

  @classmethod
  def random(cls,
             bound: 'BBox',
             rand: random.Random | None = None) -> 'Coordinate':
    """Generates a random coordinate.

    Args:
      bound: The bounding box within which the coordinate will be generated.
      rand: The random number generator to use. If None, the default random
        number generator will be used.

    Returns:
      A random coordinate.
    """
    rand = rand or random
    x = rand.randint(bound.left, bound.right)
    y = rand.randint(bound.top, bound.bottom)
    return cls(x, y)

  def distance_to(self, point: 'Coordinate') -> float:
    """Returns the distance to the point."""
    return math.sqrt((self.x - point.x) ** 2 + (self.y - point.y) ** 2)

  @classmethod
  def from_value(
      cls, value: Union[tuple[int, int], 'Coordinate']) -> 'Coordinate':
    """Creates a coordinate from a tuple or coordinate."""
    if isinstance(value, tuple):
      if len(value) == 2:
        return cls(*value)
      else:
        raise ValueError(f'Invalid tuple size: {len(value)}')
    assert isinstance(value, Coordinate), value
    return value


class BBox(pg.Object):
  """A bounding box in a 2D image."""
  x: int
  y: int
  right: int
  bottom: int

  def _on_bound(self):
    super()._on_bound()
    assert self.left < self.right, self
    assert self.top < self.bottom, self

  @property
  def left(self) -> int:
    """Returns the left coordinate of the bounding box."""
    return self.x

  @property
  def top(self) -> int:
    """Returns the top coordinate of the bounding box."""
    return self.y

  @property
  def width(self) -> int:
    """Returns the width of the bounding box."""
    return self.right - self.left

  @property
  def height(self) -> int:
    """Returns the height of the bounding box."""
    return self.bottom - self.top

  @property
  def center(self) -> Coordinate:
    """Returns the center of the bounding box."""
    return Coordinate(
        (self.left + self.right) // 2, (self.top + self.bottom) // 2)

  @property
  def area(self) -> int:
    """Returns the area of the bounding box."""
    return self.width * self.height

  @property
  def top_left(self) -> Coordinate:
    """Returns the top left corner of the bounding box."""
    return Coordinate(self.left, self.top)

  @property
  def bottom_right(self) -> Coordinate:
    """Returns the bottom right corner of the bounding box."""
    return Coordinate(self.right, self.bottom)

  def clip(self, image_size: tuple[int, int]) -> Optional['BBox']:
    """Clips the bounding box to the given image size.

    Args:
      image_size: The image size (width, height).

    Returns:
      The clipped bounding box.
    """
    width, height = image_size
    x = min(max(0, self.x), width)
    y = min(max(0, self.y), height)
    right = min(max(0, self.right), width)
    bottom = min(max(0, self.bottom), height)
    try:
      return BBox(x, y, right, bottom)
    except AssertionError:
      return None

  def __contains__(
      self,
      v: Union[
          tuple[int, int],
          tuple[int, int, int, int],
          Coordinate,
          'BBox'
      ]) -> bool:
    """Contains operator."""
    if isinstance(v, tuple):
      if len(v) == 2:
        v = Coordinate(*v)
      elif len(v) == 4:
        v = BBox(*v)
      else:
        raise ValueError(f'Invalid tuple size: {len(v)}')
    if isinstance(v, Coordinate):
      return self.left <= v.x <= self.right and self.top <= v.y <= self.bottom
    elif isinstance(v, BBox):
      return (
          v.left >= self.left
          and v.right <= self.right
          and v.top >= self.top
          and v.bottom <= self.bottom
      )
    else:
      raise ValueError(f'Invalid type: {type(v)}')

  def intersects(self, other: 'BBox') -> bool:
    """Returns the intersection of two bounding boxes."""
    return not (self.right < other.left or
                self.left > other.right or
                self.bottom < other.top or
                self.top > other.bottom)

  def as_tuple(self) -> tuple[int, int, int, int]:
    """Returns the bounding box as a tuple."""
    return self.x, self.y, self.right, self.bottom

  @classmethod
  def random(
      cls,
      bound: 'BBox',
      min_width: int = 30,
      max_width: int = 200,
      min_height: int = 30,
      max_height: int = 200,
      rand: random.Random | None = None
  ) -> 'BBox':
    """Generates a random bounding box.

    Args:
      bound: The bounding box within which the random bounding box will be
        generated.
      min_width: The minimum width of the random bounding box.
      max_width: The maximum width of the random bounding box.
      min_height: The minimum height of the random bounding box.
      max_height: The maximum height of the random bounding box.
      rand: The random number generator to use. If None, the default random
        number generator will be used.

    Returns:
      A random bounding box.
    """
    assert max_width >= min_width, (min_width, max_width)
    assert max_height >= min_height, (min_height, max_height)

    if min_width > bound.width or min_height > bound.height:
      raise ValueError('Minimum width or height is larger than the bound.')

    rand = rand or random

    max_width = min(max_width, bound.width)
    max_height = min(max_height, bound.height)

    width = rand.randint(min_width, max_width)
    height = rand.randint(min_height, max_height)

    max_left = bound.right - width
    max_top = bound.bottom - height

    left = rand.randint(bound.left, max_left)
    top = rand.randint(bound.top, max_top)

    right = left + width
    bottom = top + height

    return cls(left, top, right, bottom)

  def matches(
      self,
      other: 'BBox',
      area_diff_threshold: float = 5.0,
      max_center_distance: float = 100
  ) -> bool:
    """Returns whether the bounding boxes match."""
    if self.area == 0 or other.area == 0:
      return False

    return (
        abs(self.area - other.area) / self.area < area_diff_threshold
        and self.center.distance_to(other.center) < max_center_distance
    )

  def expand(
      self,
      width_scale: float = 1.0,
      height_scale: float = 1.0,
  ) -> 'BBox':
    """Expands the bounding box from the center."""
    new_width = int(self.width * width_scale)
    new_height = int(self.height * height_scale)
    new_x = self.x - (new_width - self.width) // 2
    new_y = self.y - (new_height - self.height) // 2
    return BBox(new_x, new_y, new_x + new_width, new_y + new_height)



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for Python code handling."""

import inspect
import unittest
from langfun.core.coding.python import parsing


class CleanTest(unittest.TestCase):

  def assert_clean(self, code: str, cleaned_code: str, clean: bool = True):
    if clean:
      cleaned_code = inspect.cleandoc(cleaned_code)
    self.assertEqual(
        parsing.clean(code), cleaned_code
    )

  def test_clean(self):
    self.assert_clean(
        """
        x = y + 1
        if x > 0:
          print(x)
        """,
        'x = y + 1\nif x > 0:\n  print(x)',
        clean=False
    )
    self.assert_clean(
        """
        def foo(x):
          return x + 1
        """,
        'def foo(x):\n  return x + 1',
        clean=False
    )
    self.assert_clean(
        """
        Here is the code:

        ```
        x = 'abc\\\''
        if len(x) > 10:
          print(x)
        ```
        """,
        """
        x = 'abc\\\''
        if len(x) > 10:
          print(x)
        """
    )
    self.assert_clean(
        """
        Here's the code:

        ```
        x = 'abc'  # Comment with '
        if len(x) > 10:
          print(x)
        ```
        """,
        """
        x = 'abc'  # Comment with '
        if len(x) > 10:
          print(x)
        """
    )
    self.assert_clean(
        """
        The code looks as below:

        ```python
        x = y + 1  # ```
        z = x * y  # "
        ```
        """,
        """
        x = y + 1  # ```
        z = x * y  # "
        """
    )
    self.assert_clean(
        """

        ```python
        x = y + 1
        '''
        Example:
          ```
          p = q + 1
          ```
        '''
        z = x * y
        ```
        """,
        """
        x = y + 1
        '''
        Example:
          ```
          p = q + 1
          ```
        '''
        z = x * y
        """
    )
    self.assert_clean(
        """
        ```python
        x = y + 1
        ```
        And another one:
        ```python
        y = z + 1
        ```
        """,
        """
        x = y + 1
        """
    )
    self.assert_clean(
        """
        ```python
        class A:
          x: int
          y: str
        ```
        """,
        """
        class A:
          x: int
          y: str
        """
    )
    self.assert_clean(
        """
        ```tool-code
        class A:
          x: int
          y: str
        ```
        """,
        """
        class A:
          x: int
          y: str
        """
    )
    self.assert_clean(
        """
        ```
        class A:
          '''Class a.

          Examples:
            ```
            A(1, 2)
            ```
          '''
          x: int
          y: str
        ```
        """,
        """
        class A:
          '''Class a.

          Examples:
            ```
            A(1, 2)
            ```
          '''
          x: int
          y: str
        """
    )

  def test_clean_with_auto_escape(self):
    self.assert_clean(
        """
        ```python
        x = 'John's home'
        ```
        """,
        """
        x = 'John\\'s home'
        """
    )
    self.assert_clean(
        """
        ```python
        x = 'Girls' home'
        ```
        """,
        """
        x = 'Girls\\' home'
        """
    )
    self.assert_clean(
        """
        ```python
        x = 'These are the girls'.'
        ```
        """,
        """
        x = 'These are the girls\\'.'
        """
    )
    self.assert_clean(
        """
        ```python
        x = 'girls'.split('')
        ```
        """,
        """
        x = 'girls'.split('')
        """
    )
    self.assert_clean(
        """
        ```python
        x = 'girls' + 'boys'
        ```
        """,
        """
        x = 'girls' + 'boys'
        """
    )
    self.assert_clean(
        """
        ```python
        x = 'girls' in ['girls', 'boys']
        ```
        """,
        """
        x = 'girls' in ['girls', 'boys']
        """
    )
    self.assert_clean(
        """
        ```python
        x = 'girls' not in ['girls', 'boys']
        ```
        """,
        """
        x = 'girls' not in ['girls', 'boys']
        """
    )
    self.assert_clean(
        """
        ```python
        x = 'Hello
        World'
        ```
        """,
        """
        x = 'Hello\\n        World'
        """
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test for langfun subscriptions."""

import dataclasses
from typing import Any
import unittest

from langfun.core import subscription


class SenderA:

  def foo(self):
    subscription.emit(FooEvent(self, 1))

  def bar(self):
    subscription.emit(BarEvent(self, 2))


@dataclasses.dataclass
class FooEvent(subscription.Event[SenderA]):
  x: int


@dataclasses.dataclass
class BarEvent(subscription.Event[SenderA]):
  y: int


@dataclasses.dataclass
class BaseEventHandler:
  events: list[subscription.Event[Any]] = dataclasses.field(
      default_factory=list
  )


class FooEventHandler(BaseEventHandler, subscription.EventHandler[FooEvent]):

  def on_event(self, event: FooEvent):
    self.events.append((self, event))


class BarEventHandler(BaseEventHandler, subscription.EventHandler[BarEvent]):

  def on_event(self, event: BarEvent):
    self.events.append((self, event))


class SenderB:

  def baz(self):
    subscription.emit(BazEvent(self, 'abc'))


@dataclasses.dataclass
class BazEvent(subscription.Event[SenderB]):
  z: str


class BazEventHandler(BaseEventHandler, subscription.EventHandler[BazEvent]):

  def on_event(self, event: BazEvent):
    self.events.append((self, event))


class EventHandlerTest(unittest.TestCase):

  def test_sender_type(self):
    self.assertIs(FooEventHandler.sender_type(), SenderA)
    self.assertIs(BarEventHandler.sender_type(), SenderA)
    self.assertIs(BazEventHandler.sender_type(), SenderB)

  def test_event_type(self):
    self.assertIs(FooEventHandler.event_type(), FooEvent)
    self.assertIs(BarEventHandler.event_type(), BarEvent)
    self.assertIs(BazEventHandler.event_type(), BazEvent)

  def test_accepts(self):
    self.assertTrue(FooEventHandler.accepts(FooEvent(SenderA(), 1)))
    self.assertFalse(FooEventHandler.accepts(BarEvent(SenderA(), 2)))
    self.assertFalse(FooEventHandler.accepts(BazEvent(SenderB(), 'abc')))

    self.assertFalse(BarEventHandler.accepts(FooEvent(SenderA(), 1)))
    self.assertTrue(BarEventHandler.accepts(BarEvent(SenderA(), 2)))
    self.assertFalse(BarEventHandler.accepts(BazEvent(SenderB(), 'abc')))

    self.assertFalse(BazEventHandler.accepts(FooEvent(SenderA(), 1)))
    self.assertFalse(BazEventHandler.accepts(BarEvent(SenderA(), 2)))
    self.assertTrue(BazEventHandler.accepts(BazEvent(SenderB(), 'abc')))


class EventTest(unittest.TestCase):

  def assertLen(self, value, length):
    self.assertEqual(len(value), length)  # pylint: disable=g-generic-assert

  def assertEmpty(self, value):
    self.assertLen(value, 0)  # pylint: disable=g-generic-assert

  def test_event_subscription_and_emition(self):
    a1 = SenderA()
    a2 = SenderA()
    foo_handler1 = FooEventHandler()
    foo_handler2 = FooEventHandler()
    bar_handler = BarEventHandler()

    b = SenderB()
    baz_handler = BazEventHandler()

    # Senders have no subscribers yet.
    self.assertEmpty(list(subscription.subscribers(a1)))
    self.assertEmpty(list(subscription.subscribers(a2)))
    self.assertEmpty(list(subscription.subscribers(b)))
    self.assertEmpty(list(subscription.subscribers(SenderA)))
    self.assertEmpty(list(subscription.subscribers(SenderB)))

    # Subscribers have not subscribed any senders yet.
    self.assertEmpty(list(subscription.subscriptions(foo_handler1)))
    self.assertEmpty(list(subscription.subscriptions(foo_handler2)))
    self.assertEmpty(list(subscription.subscriptions(bar_handler)))
    self.assertEmpty(list(subscription.subscriptions(baz_handler)))

    a1.foo()
    a1.bar()
    a2.foo()
    a2.bar()
    b.baz()

    self.assertEmpty(foo_handler1.events)
    self.assertEmpty(foo_handler2.events)
    self.assertEmpty(bar_handler.events)
    self.assertEmpty(baz_handler.events)

    # Subscribe to a sender.
    subscription.subscribe([foo_handler1, bar_handler], a1)
    subscription.subscribe(baz_handler)

    # Check subscribers.
    self.assertEqual(
        list(subscription.subscribers(a1)), [foo_handler1, bar_handler]
    )
    self.assertEmpty(list(subscription.subscribers(a2)))
    self.assertEqual(list(subscription.subscribers(b)), [baz_handler])
    self.assertEmpty(list(subscription.subscribers(SenderA)))
    self.assertEqual(list(subscription.subscribers(SenderB)), [baz_handler])

    # Check subscriptions.
    self.assertEqual(list(subscription.subscriptions(foo_handler1)), [a1])
    self.assertEmpty(list(subscription.subscriptions(foo_handler2)))
    self.assertEqual(list(subscription.subscriptions(bar_handler)), [a1])
    self.assertEqual(list(subscription.subscriptions(baz_handler)), [SenderB])

    a2.foo()
    a2.bar()
    b.baz()
    self.assertEmpty(foo_handler1.events)  # Not subscribed to a2.
    self.assertEmpty(foo_handler2.events)  # No subscriptions yet.
    self.assertEmpty(bar_handler.events)  # Not subscribe to a2.
    self.assertLen(baz_handler.events, 1)
    self.assertIs(baz_handler.events[-1][0], baz_handler)
    self.assertIs(baz_handler.events[-1][1].sender, b)

    a1.foo()
    self.assertLen(foo_handler1.events, 1)
    self.assertIs(foo_handler1.events[-1][0], foo_handler1)
    self.assertIs(foo_handler1.events[-1][1].sender, a1)
    self.assertEmpty(foo_handler2.events)  # No subscriptions yet.
    self.assertEmpty(bar_handler.events)  # a1.bar not called.
    self.assertLen(baz_handler.events, 1)  # a1.baz not called.

    a1.bar()
    self.assertLen(bar_handler.events, 1)
    self.assertIs(bar_handler.events[-1][0], bar_handler)
    self.assertIs(bar_handler.events[-1][1].sender, a1)
    self.assertLen(foo_handler1.events, 1)

    # Unsubscribe foo_handler1.
    subscription.unsubscribe(foo_handler1)
    self.assertEqual(list(subscription.subscribers(a1)), [bar_handler])
    self.assertEmpty(list(subscription.subscriptions(foo_handler1)))

    a1.foo()
    b.baz()
    self.assertLen(foo_handler1.events, 1)
    self.assertLen(baz_handler.events, 2)

    # Add subscription of a different subscriber.
    subscription.subscribe(foo_handler2, [a1, a2])
    self.assertEqual(
        list(subscription.subscribers(a1)), [bar_handler, foo_handler2]
    )
    self.assertEqual(list(subscription.subscribers(a2)), [foo_handler2])
    self.assertEqual(list(subscription.subscriptions(foo_handler2)), [a1, a2])

    a1.foo()
    a2.foo()
    b.baz()
    self.assertLen(foo_handler1.events, 1)
    self.assertLen(foo_handler2.events, 2)
    self.assertIs(foo_handler2.events[-1][0], foo_handler2)
    self.assertIs(foo_handler2.events[-1][1].sender, a2)
    self.assertLen(baz_handler.events, 3)

    # Clear all subscriptions.
    subscription.clear_subscriptions()

    # Senders have no subscribers yet.
    self.assertEmpty(list(subscription.subscribers(a1)))
    self.assertEmpty(list(subscription.subscribers(a2)))
    self.assertEmpty(list(subscription.subscribers(b)))
    self.assertEmpty(list(subscription.subscribers(SenderA)))
    self.assertEmpty(list(subscription.subscribers(SenderB)))

    # Subscribers have not subscribed any senders yet.
    self.assertEmpty(list(subscription.subscriptions(foo_handler1)))
    self.assertEmpty(list(subscription.subscriptions(foo_handler2)))
    self.assertEmpty(list(subscription.subscriptions(bar_handler)))
    self.assertEmpty(list(subscription.subscriptions(baz_handler)))

    a1.foo()
    a1.bar()
    a2.foo()
    a2.bar()
    b.baz()

    # No new events are triggered.
    self.assertLen(foo_handler1.events, 1)
    self.assertLen(foo_handler2.events, 2)
    self.assertLen(bar_handler.events, 1)
    self.assertLen(baz_handler.events, 3)

    # Bad cases.
    class A:
      pass

    with self.assertRaisesRegex(
        TypeError, '.* is not the expected sender type'):
      subscription.subscribe(foo_handler1, A())

    with self.assertRaisesRegex(TypeError, '.* is not an event handler'):
      subscription.subscribe(A())

    with self.assertRaisesRegex(
        ValueError, 'There is no subscription for'):
      subscription.unsubscribe(foo_handler1)

    with self.assertRaisesRegex(
        ValueError, 'There is no subscription for'):
      subscription.unsubscribe(foo_handler1, SenderA())


if __name__ == '__main__':
  unittest.main()
